# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

if TYPE_CHECKING:
    from synalinks.src.backend.common.variables import Variable
    from synalinks.src.embedding_models.embedding_model import EmbeddingModel


from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.backend import Trainable
from synalinks.src.backend import out_mask_json
from synalinks.src.backend.common import numpy as np
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought
from synalinks.src.optimizers.evolutionary_optimizer import EvolutionaryOptimizer
from synalinks.src.programs.program import Program
from synalinks.src.rewards.reward import squeeze_or_expand_to_same_rank
from synalinks.src.saving import serialization_lib


class MutationInputs(DataModel):
    program_description: str = Field(
        description="The program description",
    )
    program_inputs: List[Any] = Field(
        description="The inputs of the program",
    )
    program_predicted_outputs: List[Any] = Field(
        description="The program's predicted outputs",
    )
    program_ground_truth: List[Optional[Any]] = Field(
        description="The program's ground truth",
    )
    variable_description: str = Field(
        description="The description of the variable to optimize within that program"
    )
    current_variable: Any = Field(
        description="The variable to optimize",
    )


class CrossoverInputs(DataModel):
    program_description: str = Field(
        description="The program description",
    )
    program_inputs: List[Any] = Field(
        description="The inputs of the program",
    )
    program_predicted_outputs: List[Any] = Field(
        description="The program's predicted outputs",
    )
    program_ground_truth: List[Optional[Any]] = Field(
        description="The program's ground truth",
    )
    variable_description: str = Field(
        description="The description of the variable to optimize within that program",
    )
    other_variable: Any = Field(
        description="other high performing variable to merge",
    )
    current_variable: Any = Field(
        description="current high performing variable to merge",
    )


def base_instructions():
    """Base instructions that define the context for all optimization programs.

    These instructions explain that the system optimizes JSON variables
    in a computation graph.
    """
    return """
You are an integral part of an optimization system designed to improve
JSON variables within a computation graph (i.e. the program).
Each module in the graph performs specific computations, with JSON variables
serving as the state.
These variables can represent prompts, code, plans, rules, or any other
JSON-compatible data.
""".strip()


def mutation_instructions(variables_keys):
    """Instructions for the mutation program that optimizes variables.

    Args:
        variables_keys (list): List of keys that the variable should contain
    """
    return f"""
Your primary task is to creatively enhance the provided variable so that the
predicted output aligns as closely as possible with the ground truth.
Pay close attention to the variable's description, its intended use, and the
broader context of the computation graph.

Guidelines:
- Ensure the new variable is generalizable and performs well across various
  inputs of the same kind.
- Include all specified keys: {variables_keys}.
- Justify each change with clear reasoning, referencing the variable's purpose
  and the desired output.
- If no ground truth is provided, the goal is to critically enhance the
  predicted output.
- If you have to optimize a variable containing code, provide a generalizable
  algorithm.
- Always focus on ONLY one aspect at the time.
- If the instructions/prompt contains general information, keep it.
""".strip()


def crossover_instructions(variables_keys):
    """Instructions for the crossover program that optimizes variables.

    Args:
        variables_keys (list): List of keys that the variable should contain
    """
    return f"""
Your responsibility is to create a new, optimized variable by strategically
combining features from the current variable and a high-performing candidate.
The new variable should improve the alignment of the predicted output with
the ground truth.

Guidelines:
- Analyze both the current variable and the other high-performing variable,
  identifying their respective strengths and weaknesses.
- Pay close attention to the variable's description, its intended use, and the
  broader context of the computation graph.
- Ensure the new variable is generalizable and performs well across various
  inputs of the same kind.
- Include all specified keys: {variables_keys}.
- Justify each feature you incorporate, explaining how it contributes to
  better performance or alignment with the ground truth.
- If no ground truth is provided, the goal is to critically enhance the
  predicted output.
- If you have to optimize a variable containing code, provide a generalizable
  algorithm.
- Always focus on ONLY one aspect at the time.
- If the instructions/prompt contains general information, keep it.
""".strip()


async def similarity_distance(
    candidate1: Dict[str, Any],
    candidate2: Dict[str, Any],
    embedding_model: Optional["EmbeddingModel"] = None,
    axis: int = -1,
) -> float:
    """Compute distance between two candidates using embeddings.

    This function computes the cosine distance between the mean embeddings
    of two candidate JSON objects. Each field of the JSON is embedded
    separately, normalized to unit length, then averaged.

    Args:
        candidate1 (dict): First candidate (dict or JSON-serializable object)
        candidate2 (dict): Second candidate (dict or JSON-serializable object)
        embedding_model (EmbeddingModel): The embedding model for computing embeddings
        axis (int): The axis along which to compute the similarity (default: -1)

    Returns:
        float: Cosine distance between candidates (0 = identical, 1 = orthogonal)
    """
    embeddings1 = await embedding_model(tree.flatten(candidate1))
    embeddings2 = await embedding_model(tree.flatten(candidate2))
    embeddings1 = embeddings1["embeddings"]
    embeddings2 = embeddings2["embeddings"]
    embeddings1 = np.convert_to_tensor(embeddings1)
    embeddings2 = np.convert_to_tensor(embeddings2)
    embeddings1, embeddings2 = squeeze_or_expand_to_same_rank(embeddings1, embeddings2)
    embeddings1 = np.normalize(embeddings1, axis=axis)
    embeddings2 = np.normalize(embeddings2, axis=axis)
    embeddings1 = np.mean(embeddings1, axis=0)
    embeddings2 = np.mean(embeddings2, axis=0)
    similarity = (np.sum(embeddings1 * embeddings2, axis=axis) + 1) / 2
    return 1 - similarity


@synalinks_export("synalinks.optimizers.OMEGA")
class OMEGA(EvolutionaryOptimizer):
    """OMEGA: OptiMizEr as Genetic Algorithm.

    A genetic optimizer with dominated novelty search.

    This optimizer is **unique to Synalinks** and the result of our research
    effort on advancing neuro-symbolic AI.

    Dominated Novelty Search (DNS), is a SOTA Quality-Diversity optimization
    method that implements a competition function in a classic genetic
    algorithm.

    The key insight behind Dominated Novelty Search is that candidates should
    be eliminated from the population if they are both:

    - Inferior in reward/fitness
    - Similar to existing candidates/solutions

    This algorithm creates an evolutionary pressure to focus on high performing
    candidates **Or** candidates that explore other approaches.

    This approach only add one step to the traditional genetic algorithm and
    *outperform* MAP-Elites, Threshold-Elites and Cluster-Elites.

    This allow the system to explore the search space more quickly by
    eliminating non-promising candidates while preserving diversity to avoid
    local optimum.

    At Synalinks, we adapted this algorithm for LM-based optimization, to do
    so we use an embedding model to compute the candidate's descriptor and a
    cosine distance between solutions.

    **Note**: In Synalinks, unlike other In-Context learning frameworks, a
    variable (the module's state to optimize) is a JSON object not a simple
    string. Which has multiple implications, we maintain a 100% correct
    structure through constrained JSON decoding, and we allow the state to
    have variable/dynamic number of fields, which is handled by this approach
    by embedding each field and averaging them before computing the distance
    required by DNS.

    Example:
    ```
    import synalinks
    import asyncio

    async def main():
        # ... your program definition

        program.compile(
            reward=synalinks.rewards.ExactMatch(),
            optimizer=synalinks.optimizers.OMEGA(
                language_model=language_model,
                embedding_model=embedding_model,
            )
        )

        history = await program.fit(...)
    ```

    Concerning the inspirations for this optimizer:
        - Dominated Novelty Search for their elegant Quality-Diversity
          algorithm that outperform many other evolutionary strategies.
        - DSPY's GEPA for feeding the optimizer program with the raw training
          data and for formalizing the evolutionary optimization strategy
          (**NOT** the MAP-Elites method used).
        - DeepMind's AlphaEvolve have been a huge inspiration, more on the
          motivational side as they didn't released the code.

    References:
        - Dominated Novelty Search: Rethinking Local Competition in
          Quality-Diversity (https://arxiv.org/html/2502.00593v1)
        - GEPA: Reflective Prompt Evolution Can Outperform Reinforcement
          Learning (https://arxiv.org/pdf/2507.19457)
        - AlphaEvolve: A coding agent for scientific and algorithmic
          discovery (https://arxiv.org/pdf/2506.13131)

    Args:
        instructions (str): Additional instructions about the task for the
            optimizer.
        language_model (LanguageModel): The language model to use.
        embedding_model (EmbeddingModel): The embedding model to use to
            compute candidates descriptors according to Dominated Novelty
            Search.
        k_nearest_fitter (int): The K nearest fitter used by Dominated
            Novelty Search.
        distance_function (callable): Optional. The distance function to use
            by Dominated Novelty Search. If no function is provided, use
            the default cosine distance.
        mutation_temperature (float): The temperature for the LM calls of
            the mutation programs.
        crossover_temperature (float): The temperature for the LM calls of
            the crossover programs.
        reasoning_effort (string): Optional. The reasoning effort for the LM call
            between ['minimal', 'low', 'medium', 'high', 'disable', 'none', None].
            Default to None (no reasoning).
        algorithm (str): The mechanism to use for the genetic algorithm
            between ['ga', 'dns']. This parameter is provided for ablation
            studies and shouldn't be modified. (Default to 'dns').
        selection (str): The method to select the candidate to evolve at the
            beginning of a batch between ['random', 'best', 'softmax'].
            (Default to 'softmax').
        selection_temperature (float): The temperature for softmax selection.
            Used only when `selection='softmax'`. Lower values concentrate
            selection on high-reward candidates, higher values make selection
            more uniform (Default 0.3).
        merging_rate (float): Rate at which crossover vs mutation is selected.
            (Default to 0.02).
        population_size (int): The maximum number of best candidates to keep
            during the optimization process.
        name (str): Optional name for the optimizer instance.
        description (str): Optional description of the optimizer instance.
    """

    def __init__(
        self,
        instructions=None,
        language_model=None,
        embedding_model=None,
        k_nearest_fitter=5,
        distance_function=None,
        mutation_temperature=0.3,
        crossover_temperature=0.3,
        reasoning_effort=None,
        merging_rate=0.02,
        algorithm="dns",
        selection="softmax",
        selection_temperature=0.3,
        population_size=10,
        name=None,
        description=None,
        **kwargs,
    ):
        super().__init__(
            language_model=language_model,
            mutation_temperature=mutation_temperature,
            crossover_temperature=crossover_temperature,
            selection=selection,
            selection_temperature=selection_temperature,
            merging_rate=merging_rate,
            population_size=population_size,
            name=name,
            description=description,
            **kwargs,
        )
        if not instructions:
            instructions = ""
        self.instructions = instructions
        self.reasoning_effort = reasoning_effort

        # DNS-specific parameters
        self.embedding_model = embedding_model
        self.k_nearest_fitter = k_nearest_fitter
        self.distance_function = distance_function

        algorithms = ["ga", "dns"]
        if algorithm not in algorithms:
            raise ValueError(f"Parameter `algorithm` should be between {algorithms}")
        self.algorithm = algorithm

    async def build(self, trainable_variables):
        """
        Build the optimizer programs based on the trainable variables.

        Args:
            trainable_variables (list): List of variables that will be optimized
        """
        for trainable_variable in trainable_variables:
            schema_id = id(trainable_variable.get_schema())
            mask = list(Trainable.keys())
            symbolic_variable = trainable_variable.to_symbolic_data_model().out_mask(
                mask=mask
            )

            if schema_id not in self.mutation_programs:
                inputs = Input(data_model=MutationInputs)
                outputs = await ChainOfThought(
                    data_model=symbolic_variable,
                    language_model=self.language_model,
                    temperature=self.mutation_temperature,
                    reasoning_effort=self.reasoning_effort,
                    instructions=(
                        "\n".join(
                            [
                                base_instructions(),
                                mutation_instructions(list(symbolic_variable.keys())),
                            ]
                        )
                        if not self.instructions
                        else "\n".join(
                            [
                                self.instructions,
                                base_instructions(),
                                mutation_instructions(list(symbolic_variable.keys())),
                            ]
                        )
                    ),
                    name=f"mutation_cot_{schema_id}_" + self.name,
                )(inputs)
                outputs = outputs.in_mask(mask=list(symbolic_variable.keys()))
                program = Program(
                    inputs=inputs,
                    outputs=outputs,
                    name=f"mutation_{schema_id}_" + self.name,
                    description="The mutation program that fix/optimize variables",
                )
                self.mutation_programs[schema_id] = program

            if schema_id not in self.crossover_programs:
                inputs = Input(data_model=CrossoverInputs)
                outputs = await ChainOfThought(
                    data_model=symbolic_variable,
                    language_model=self.language_model,
                    temperature=self.crossover_temperature,
                    reasoning_effort=self.reasoning_effort,
                    instructions=(
                        "\n".join(
                            [
                                base_instructions(),
                                crossover_instructions(list(symbolic_variable.keys())),
                            ]
                        )
                        if not self.instructions
                        else "\n".join(
                            [
                                self.instructions,
                                base_instructions(),
                                crossover_instructions(list(symbolic_variable.keys())),
                            ]
                        )
                    ),
                    name=f"crossover_cot_{schema_id}_" + self.name,
                )(inputs)
                outputs = outputs.in_mask(mask=list(symbolic_variable.keys()))
                program = Program(
                    inputs=inputs,
                    outputs=outputs,
                    name=f"crossover_{schema_id}_" + self.name,
                    description="Crossover program combining high performing variables",
                )
                self.crossover_programs[schema_id] = program

        self.built = True

    async def mutate_candidate(
        self,
        step: int,
        trainable_variable: "Variable",
        selected_candidate: Dict[str, Any],
        x: Optional[List[Any]] = None,
        y: Optional[List[Any]] = None,
        y_pred: Optional[List[Any]] = None,
        training: bool = False,
    ) -> Dict[str, Any]:
        """Apply mutation to generate a new candidate using LLM.

        Creates mutation inputs from the selected candidate and training data,
        then calls the mutation program to generate an optimized variant.

        Args:
            step (int): The current training step
            trainable_variable (Variable): The trainable variable (for metadata access)
            selected_candidate (dict): The selected candidate to mutate
            x (list): Input data batch
            y (list): Ground truth data batch
            y_pred (list): Predicted outputs from the current model
            training (bool): Whether in training mode

        Returns:
            dict: The mutated candidate from the mutation program
        """
        mask = list(Trainable.keys())
        schema_id = id(trainable_variable.get_schema())
        masked_variable = out_mask_json(
            selected_candidate,
            mask=mask,
        )
        inputs = MutationInputs(
            program_description=self.program.description,
            program_inputs=[inp.get_json() for inp in x],
            program_predicted_outputs=[
                pred.get_json() if pred else None for pred in y_pred
            ],
            program_ground_truth=([gt.get_json() for gt in y] if y is not None else []),
            variable_description=trainable_variable.description,
            current_variable=masked_variable,
        )
        program = self.mutation_programs[schema_id]
        return await program(inputs, training=training)

    async def merge_candidate(
        self,
        step: int,
        trainable_variable: "Variable",
        current_candidate: Dict[str, Any],
        other_candidate: Dict[str, Any],
        x: Optional[List[Any]] = None,
        y: Optional[List[Any]] = None,
        y_pred: Optional[List[Any]] = None,
        training: bool = False,
    ) -> Dict[str, Any]:
        """Apply crossover to merge two selected candidates.

        Creates crossover inputs combining two high-performing candidates,
        then calls the crossover program to generate a merged variant.

        Args:
            step (int): The current training step
            trainable_variable (Variable): The trainable variable (for metadata access)
            current_candidate (dict): First selected candidate to merge
            other_candidate (dict): Second selected candidate to merge
            x (list): Input data batch
            y (list): Ground truth data batch
            y_pred (list): Predicted outputs from the current model
            training (bool): Whether in training mode

        Returns:
            dict: The merged candidate from the crossover program
        """
        mask = list(Trainable.keys())
        schema_id = id(trainable_variable.get_schema())
        current_variable = out_mask_json(
            current_candidate,
            mask=mask,
        )
        other_variable = out_mask_json(
            other_candidate,
            mask=mask,
        )
        inputs = CrossoverInputs(
            program_description=self.program.description,
            program_inputs=[inp.get_json() for inp in x],
            program_predicted_outputs=[
                pred.get_json() if pred else None for pred in y_pred
            ],
            program_ground_truth=([gt.get_json() for gt in y] if y is not None else []),
            variable_description=trainable_variable.description,
            other_variable=other_variable,
            current_variable=current_variable,
        )
        program = self.crossover_programs[schema_id]
        return await program(inputs, training=training)

    async def competition(self, candidates: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Apply Dominated Novelty Search (DNS) competition.

        DNS filters candidates by removing those that are both:
        - Inferior in reward (dominated)
        - Similar to existing candidates (not novel)

        This maintains diversity while focusing on high-performing candidates.

        Args:
            candidates (list): List of candidate dictionaries with 'reward' key

        Returns:
            list: Filtered list of candidates that passed the DNS competition
        """
        if len(candidates) <= 1:
            return candidates

        distance_function = (
            self.distance_function if self.distance_function else similarity_distance
        )

        selected_candidates = []
        for candidate in candidates:
            is_dominated = False
            for other in candidates:
                if other is candidate:
                    continue
                distance = await distance_function(
                    candidate,
                    other,
                    embedding_model=self.embedding_model,
                )
                # Check if within k-nearest neighborhood
                if distance < 1.0 / self.k_nearest_fitter:
                    # Check if dominated (lower reward)
                    if candidate.get("reward", 0) < other.get("reward", 0):
                        is_dominated = True
                        break
            if not is_dominated:
                selected_candidates.append(candidate)

        return selected_candidates if selected_candidates else [candidates[0]]

    async def on_epoch_end(self, epoch, trainable_variables):
        """Called at the end of each epoch.

        Applies DNS competition (if algorithm='dns') to filter candidates,
        then selects the top candidates based on population_size.

        Args:
            epoch (int): The epoch number
            trainable_variables (list): The list of trainable variables
        """
        for trainable_variable in trainable_variables:
            candidates = trainable_variable.get("candidates")
            best_candidates = trainable_variable.get("best_candidates")

            # Combine current candidates with best candidates
            all_candidates = candidates + best_candidates

            # Apply DNS competition if enabled
            if self.algorithm == "dns" and len(all_candidates) > 1:
                all_candidates = await self.competition(all_candidates)

            # Sort by reward and keep top population_size candidates
            all_candidates = sorted(
                all_candidates,
                key=lambda x: x.get("reward", 0),
                reverse=True,
            )
            trainable_variable.update(
                {
                    "candidates": [],
                    "best_candidates": all_candidates[: self.population_size],
                }
            )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "instructions": self.instructions,
                "reasoning_effort": self.reasoning_effort,
                "k_nearest_fitter": self.k_nearest_fitter,
                "algorithm": self.algorithm,
            }
        )
        if self.embedding_model:
            config["embedding_model"] = serialization_lib.serialize_synalinks_object(
                self.embedding_model
            )
        return config

    @classmethod
    def from_config(cls, config):
        embedding_model = None
        if "embedding_model" in config:
            embedding_model = serialization_lib.deserialize_synalinks_object(
                config.pop("embedding_model")
            )
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model")
        )
        return cls(
            language_model=language_model,
            embedding_model=embedding_model,
            **config,
        )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import os
import tempfile
from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.backend import JsonDataModel
from synalinks.src.knowledge_bases import KnowledgeBase
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules.knowledge.retrieve_knowledge import RetrieveKnowledge


class Document(DataModel):
    id: str = Field(description="The document id")
    text: str = Field(description="The content of the document")


class Query(DataModel):
    question: str = Field(description="The user question")


class RetrieveKnowledgeTest(testing.TestCase):
    def setUp(self):
        super().setUp()
        self.temp_dir = tempfile.mkdtemp()
        self.db_path = os.path.join(self.temp_dir, "test.db")

    def tearDown(self):
        import shutil

        shutil.rmtree(self.temp_dir, ignore_errors=True)
        super().tearDown()

    @patch("litellm.acompletion")
    async def test_retrieve_knowledge(self, mock_completion):
        # Mock the LLM response to return a search query
        mock_completion.return_value = {
            "choices": [
                {
                    "message": {
                        "content": (
                            '{"tables": ["Document"], "search": ["quick brown fox"]}'
                        )
                    }
                }
            ]
        }

        knowledge_base = KnowledgeBase(
            uri="duckdb://" + self.db_path,
            data_models=[Document],
        )

        # Store some documents
        docs = [
            JsonDataModel(
                data_model=Document(id="doc1", text="The quick brown fox jumps")
            ),
            JsonDataModel(data_model=Document(id="doc2", text="The lazy dog sleeps")),
        ]
        await knowledge_base.update(docs)

        language_model = LanguageModel(model="openai/gpt-4o-mini")

        retrieve_module = RetrieveKnowledge(
            knowledge_base=knowledge_base,
            language_model=language_model,
            data_models=[Document.to_symbolic_data_model()],
            k=5,
            name="test_retriever",
        )

        query = JsonDataModel(data_model=Query(question="Find documents about foxes"))
        result = await retrieve_module(query)

        self.assertIsNotNone(result)
        self.assertIn("result", result.get_json())

    async def test_retrieve_knowledge_none_input(self):
        knowledge_base = KnowledgeBase(
            uri=self.db_path,
            data_models=[Document],
        )

        language_model = LanguageModel(model="openai/gpt-4o-mini")

        retrieve_module = RetrieveKnowledge(
            knowledge_base=knowledge_base,
            language_model=language_model,
            data_models=[Document.to_symbolic_data_model()],
            name="test_retriever",
        )

        result = await retrieve_module(None)
        self.assertIsNone(result)

    def test_retrieve_knowledge_default_instructions(self):
        knowledge_base = KnowledgeBase(
            uri=self.db_path,
            data_models=[Document],
        )

        language_model = LanguageModel(model="openai/gpt-4o-mini")

        retrieve_module = RetrieveKnowledge(
            knowledge_base=knowledge_base,
            language_model=language_model,
            data_models=[Document.to_symbolic_data_model()],
            name="test_retriever",
        )

        self.assertIn("Document", retrieve_module.instructions)
        self.assertIn("search", retrieve_module.instructions)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import asyncio
import logging
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend import any_symbolic_data_models
from synalinks.src.backend.config import mlflow_experiment_name
from synalinks.src.backend.config import mlflow_tracking_uri
from synalinks.src.hooks.hook import Hook
from synalinks.src.utils.async_utils import run_maybe_nested

try:
    import mlflow
    from mlflow.entities import SpanType

    MLFLOW_AVAILABLE = True
except ImportError:
    MLFLOW_AVAILABLE = False
    SpanType = None

# Global registry to track spans across hook instances for parent-child relationships
_GLOBAL_SPANS_REGISTRY: Dict[str, Any] = {}


@synalinks_export("synalinks.callbacks.monitor.Span")
class Span(DataModel):
    """Data model representing a span for module call tracing."""

    event: Literal["call_begin", "call_end"]
    is_symbolic: bool
    call_id: str
    parent_call_id: Optional[str]
    module: str
    module_name: str
    module_description: str
    timestamp: float
    inputs: Optional[List[Dict[str, Any]]] = None
    outputs: Optional[List[Dict[str, Any]]] = None
    duration: Optional[float] = None
    exception: Optional[str] = None
    success: Optional[bool] = None
    cost: Optional[float] = None


@synalinks_export("synalinks.hooks.Monitor")
class Monitor(Hook):
    """Monitor hook for tracing module calls using MLflow.

    This hook creates MLflow spans for each module call, enabling distributed
    tracing and observability of your synalinks programs.

    You can enable monitoring for every module by using
    `synalinks.enable_observability()` at the beginning of your scripts.

    Args:
        tracking_uri (str): MLflow tracking server URI. If None, uses the
            value from `synalinks.enable_observability()` or the default
            (local ./mlruns directory or MLFLOW_TRACKING_URI env var).
        experiment_name (str): Name of the MLflow experiment for tracing.
            If None, uses the value from `synalinks.enable_observability()`
            or defaults to "synalinks_traces".

    Example:

    ```python
    import synalinks

    # Basic usage - uses local MLflow storage
    synalinks.enable_observability()

    # With custom MLflow tracking server
    synalinks.enable_observability(
        tracking_uri="http://localhost:5000",
        experiment_name="my_traces"
    )

    # Or create a Monitor hook directly with custom settings
    monitor = synalinks.hooks.Monitor(
        tracking_uri="http://localhost:5000",
        experiment_name="my_experiment"
    )
    ```
    """

    def __init__(
        self,
        tracking_uri=None,
        experiment_name=None,
    ):
        super().__init__()
        if not MLFLOW_AVAILABLE:
            raise ImportError(
                "mlflow is required for the Monitor hook. "
                "Install it with: pip install mlflow"
            )

        # Use provided values or fall back to global config
        self.tracking_uri = tracking_uri or mlflow_tracking_uri()
        self.experiment_name = experiment_name or mlflow_experiment_name()
        self.call_start_times = {}
        self._spans = {}
        self.logger = logging.getLogger(__name__)
        self._setup_done = False

    def _setup_mlflow(self):
        """Configure MLflow tracking."""
        if self._setup_done:
            return

        if self.tracking_uri:
            mlflow.set_tracking_uri(self.tracking_uri)

        mlflow.set_experiment(self.experiment_name)
        self._setup_done = True

    def _serialize_data(self, data):
        """Serialize data models to JSON-compatible format."""
        flatten_data = tree.flatten(data)
        is_symbolic = any_symbolic_data_models(data)

        if is_symbolic:
            serialized = [dm.get_schema() for dm in flatten_data if dm is not None]
        else:
            serialized = [dm.get_json() for dm in flatten_data if dm is not None]

        return serialized, is_symbolic

    def _get_span_type(self):
        """Determine the MLflow span type based on the module class."""
        if SpanType is None:
            return None

        module_class = self.module.__class__.__name__

        # Map module types to MLflow span types
        if module_class in ("Generator", "ChainOfThought", "SelfCritique"):
            return SpanType.LLM
        elif module_class in ("FunctionCallingAgent",):
            return SpanType.AGENT
        elif module_class in ("EmbedKnowledge", "RetrieveKnowledge", "UpdateKnowledge"):
            return SpanType.RETRIEVER
        elif module_class in ("Tool",):
            return SpanType.TOOL
        else:
            return SpanType.CHAIN

    async def _begin_span_async(
        self,
        call_id,
        parent_call_id,
        serialized_inputs,
        serialized_kwargs,
        is_symbolic,
        span_name,
        span_type,
    ):
        """Async implementation of span creation."""
        global _GLOBAL_SPANS_REGISTRY

        # Look up parent span from global registry for proper trace hierarchy
        parent_span_obj = None
        if parent_call_id and parent_call_id in _GLOBAL_SPANS_REGISTRY:
            parent_span_obj = _GLOBAL_SPANS_REGISTRY[parent_call_id]

        # Use start_span_no_context for manual lifecycle management
        # This properly supports parent-child relationships
        span = await asyncio.to_thread(
            mlflow.start_span_no_context,
            name=span_name,
            span_type=span_type,
            parent_span=parent_span_obj,
        )

        # Store in both local and global registry
        self._spans[call_id] = span
        _GLOBAL_SPANS_REGISTRY[call_id] = span

        span.set_attributes(
            {
                "synalinks.call_id": call_id,
                "synalinks.parent_call_id": parent_call_id or "",
                "synalinks.module": str(self.module.__class__.__name__),
                "synalinks.module_name": self.module.name or "",
                "synalinks.module_description": self.module.description or "",
                "synalinks.is_symbolic": is_symbolic,
            }
        )

        # Set inputs as a dictionary (MLflow handles serialization)
        inputs_dict = {"data": serialized_inputs}
        if serialized_kwargs:
            inputs_dict["kwargs"] = serialized_kwargs
        span.set_inputs(inputs_dict)

        self.logger.debug(f"Started span for call {call_id}: {span_name}")

    def on_call_begin(
        self,
        call_id,
        parent_call_id=None,
        inputs=None,
        kwargs=None,
    ):
        """Called when a module call begins."""
        self._setup_mlflow()
        self.call_start_times[call_id] = time.time()

        serialized_inputs, is_symbolic = self._serialize_data(inputs)

        # Serialize kwargs if present (for modules that use keyword arguments)
        serialized_kwargs = {}
        if kwargs:
            # Filter out non-serializable kwargs like 'training'
            for key, value in kwargs.items():
                if key == "training":
                    serialized_kwargs[key] = value
                elif hasattr(value, "get_json"):
                    serialized_kwargs[key] = value.get_json()
                elif isinstance(value, (str, int, float, bool, list, dict, type(None))):
                    serialized_kwargs[key] = value

        span_name = f"{self.module.__class__.__name__}"
        if self.module.name:
            span_name = f"{span_name}:{self.module.name}"

        # Get the appropriate span type for this module
        span_type = self._get_span_type()

        run_maybe_nested(
            self._begin_span_async(
                call_id=call_id,
                parent_call_id=parent_call_id,
                serialized_inputs=serialized_inputs,
                serialized_kwargs=serialized_kwargs,
                is_symbolic=is_symbolic,
                span_name=span_name,
                span_type=span_type,
            )
        )

    async def _end_span_async(
        self,
        call_id,
        span,
        serialized_outputs,
        duration,
        cost,
        exception,
    ):
        """Async implementation of span ending."""
        span.set_attributes(
            {
                "synalinks.duration": duration,
                "synalinks.success": exception is None,
                "synalinks.cost": cost or 0.0,
            }
        )

        if exception:
            span.set_attributes({"synalinks.exception": str(exception)})
            # Add exception event for better visibility in MLflow UI
            span.add_event(
                mlflow.entities.SpanEvent(
                    name="exception",
                    attributes={
                        "exception.type": type(exception).__name__,
                        "exception.message": str(exception),
                    },
                )
            )
            span.set_status("ERROR")
        else:
            span.set_status("OK")

        # Set outputs as a dictionary (MLflow handles serialization)
        span.set_outputs({"data": serialized_outputs})

        await asyncio.to_thread(span.end)

        success = exception is None
        self.logger.debug(
            f"Ended span for call {call_id}, duration={duration:.3f}s, success={success}"
        )

    def on_call_end(
        self,
        call_id,
        parent_call_id=None,
        outputs=None,
        exception=None,
    ):
        """Called when a module call ends."""
        global _GLOBAL_SPANS_REGISTRY

        end_time = time.time()
        start_time = self.call_start_times.pop(call_id, end_time)
        duration = end_time - start_time

        span = self._spans.pop(call_id, None)
        # Also remove from global registry
        _GLOBAL_SPANS_REGISTRY.pop(call_id, None)

        if span is None:
            self.logger.warning(f"No span found for call_id {call_id}")
            return

        serialized_outputs, _ = self._serialize_data(outputs)

        cost = None
        if self.module._get_call_context():
            cost = self.module._get_call_context().cost

        run_maybe_nested(
            self._end_span_async(
                call_id=call_id,
                span=span,
                serialized_outputs=serialized_outputs,
                duration=duration,
                cost=cost,
                exception=exception,
            )
        )

    def __del__(self):
        """Cleanup any open spans."""
        if hasattr(self, "_spans"):
            for call_id, span in list(self._spans.items()):
                try:
                    span.end()
                except Exception:
                    pass
            self._spans.clear()



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend.common.json_data_model import JsonDataModel


class JsonDataModelTest(testing.TestCase):
    def test_init_with_data_model(self):
        class Query(DataModel):
            query: str

        json_data_model = JsonDataModel(
            data_model=Query(query="What is the capital of France?")
        )

        expected_schema = {
            "additionalProperties": False,
            "properties": {"query": {"title": "Query", "type": "string"}},
            "required": ["query"],
            "title": "Query",
            "type": "object",
        }
        expected_json = {"query": "What is the capital of France?"}

        self.assertEqual(json_data_model.get_json(), expected_json)
        self.assertEqual(json_data_model.get_schema(), expected_schema)

    def test_init_with_data_model_non_instanciated(self):
        class Query(DataModel):
            query: str

        with self.assertRaisesRegex(ValueError, "Couldn't get the JSON data"):
            _ = JsonDataModel(data_model=Query)

    def test_init_with_data_model_non_instanciated_and_value(self):
        class Query(DataModel):
            query: str

        expected_schema = {
            "additionalProperties": False,
            "properties": {"query": {"title": "Query", "type": "string"}},
            "required": ["query"],
            "title": "Query",
            "type": "object",
        }
        expected_json = {"query": "What is the capital of France?"}

        json_data_model = JsonDataModel(
            data_model=Query,
            json=expected_json,
        )

        self.assertEqual(json_data_model.get_json(), expected_json)
        self.assertEqual(json_data_model.get_schema(), expected_schema)

    def test_init_with_dict(self):
        schema = {
            "additionalProperties": False,
            "properties": {"query": {"title": "Query", "type": "string"}},
            "required": ["query"],
            "title": "Query",
            "type": "object",
        }
        value = {"query": "What is the capital of France?"}

        json_data_model = JsonDataModel(schema=schema, json=value)

        self.assertEqual(json_data_model.get_json(), value)
        self.assertEqual(json_data_model.get_schema(), schema)

    def test_representation(self):
        class Query(DataModel):
            query: str

        json_data_model = JsonDataModel(
            data_model=Query(query="What is the capital of France?")
        )

        expected_schema = {
            "additionalProperties": False,
            "properties": {"query": {"title": "Query", "type": "string"}},
            "required": ["query"],
            "title": "Query",
            "type": "object",
        }
        expected_json = {"query": "What is the capital of France?"}

        self.assertEqual(
            str(json_data_model),
            f"<JsonDataModel schema={expected_schema}, json={expected_json}>",
        )

    def test_contains_json_data_model(self):
        class Foo(DataModel):
            foo: str

        class FooBar(DataModel):
            foo: str
            bar: str

        class Bar(DataModel):
            bar: str

        foo_json = JsonDataModel(data_model=Foo(foo="a"))
        foobar_json = JsonDataModel(data_model=FooBar(foo="a", bar="b"))
        bar_json = JsonDataModel(data_model=Bar(bar="c"))

        self.assertTrue(foo_json in foobar_json)
        self.assertFalse(bar_json in foo_json)

    def test_contains_string_key_json_data_model(self):
        class FooBar(DataModel):
            foo: str
            bar: str

        foobar_json = JsonDataModel(data_model=FooBar(foo="a", bar="b"))

        self.assertTrue("foo" in foobar_json)
        self.assertTrue("bar" in foobar_json)
        self.assertFalse("baz" in foobar_json)

    def test_not_json_data_model(self):
        class Foo(DataModel):
            foo: str

        foo_json = JsonDataModel(data_model=Foo(foo="a"))

        not_foo = ~foo_json

        self.assertTrue(not_foo is None)



import inspect
import typing
from typing import cast

from mcp import ClientSession
from mcp.types import CallToolResult
from mcp.types import EmbeddedResource
from mcp.types import ImageContent
from mcp.types import TextContent
from mcp.types import Tool as MCPTool

from synalinks.src.modules.core.tool import Tool
from synalinks.src.utils.mcp.sessions import Connection
from synalinks.src.utils.mcp.sessions import create_session

NonTextContent = ImageContent | EmbeddedResource
MAX_ITERATIONS = 1000


class ToolException(Exception):  # noqa: N818
    """Exception thrown when a tool execution error occurs.

    This exception allows tools to signal errors without stopping the agent.
    """


class ToolMessage(typing.TypedDict):
    """The tool response JSON schema"""

    response: str | list[str]


# TODO: support for non-text artifacts in the tool response
def _convert_call_tool_result(
    call_tool_result: CallToolResult,
) -> tuple[str | list[str], list[NonTextContent] | None]:
    text_contents: list[TextContent] = []
    non_text_contents = []

    for content in call_tool_result.content:
        if isinstance(content, TextContent):
            text_contents.append(content)
        else:
            non_text_contents.append(content)

    tool_content: str | list[str] = [content.text for content in text_contents]
    if not text_contents:
        tool_content = ""
    elif len(text_contents) == 1:
        tool_content = tool_content[0]

    if call_tool_result.isError:
        raise ToolException(tool_content)

    tool_message = {
        "response": tool_content,
        # "artifact": non_text_contents,
    }

    return tool_message


async def _list_all_tools(session: ClientSession) -> list[MCPTool]:
    current_cursor: str | None = None
    all_tools: list[MCPTool] = []

    iterations = 0

    while True:
        iterations += 1
        if iterations > MAX_ITERATIONS:
            raise RuntimeError("Reached max of 1000 iterations while listing tools.")

        list_tools_page_result = await session.list_tools(cursor=current_cursor)

        if list_tools_page_result.tools:
            all_tools.extend(list_tools_page_result.tools)

        if list_tools_page_result.nextCursor is None:
            break

        current_cursor = list_tools_page_result.nextCursor
    return all_tools


def _has_docstring_section(docstring: str, section: str) -> bool:
    """Check if a docstring already contains a specific section (Args, Returns, etc.)."""
    if not docstring:
        return False

    # Look for common docstring section patterns
    section_patterns = [
        f"\n{section}:",
        f"\n{section.lower()}:",
        f"\n{section.upper()}:",
        f"\n{section.capitalize()}:",
    ]

    return any(pattern in docstring for pattern in section_patterns)


def _create_async_function_from_mcp_tool(
    mcp_tool: MCPTool,
    session: ClientSession | None,
    connection: Connection | None = None,
    namespace: str | None = None,
) -> typing.Coroutine:
    """Create a dynamic async function from an MCP tool
    that can be wrapped by Synalinks tool.

    """
    properties = (
        mcp_tool.inputSchema.get("properties", {}) if mcp_tool.inputSchema else {}
    )
    required_params = (
        set(mcp_tool.inputSchema.get("required", [])) if mcp_tool.inputSchema else set()
    )

    parameters = []
    annotations = {}

    for param_name, param_schema in properties.items():
        param_type = _json_schema_to_python_type(param_schema)
        annotations[param_name] = param_type

        if param_name in required_params:
            param = inspect.Parameter(
                param_name, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=param_type
            )
        else:
            param = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=param_type,
                default=None,
            )

        parameters.append(param)

    name = f"{namespace}_{mcp_tool.name}" if namespace else mcp_tool.name

    description = mcp_tool.description or f"Execute {name} function"
    docstring = [description]

    if properties and not _has_docstring_section(description, "Args"):
        docstring.append("\nArgs:")

        for param_name, param_schema in properties.items():
            param_desc = param_schema.get("description", f"{param_name} parameter")
            docstring.append(f"    {param_name}: {param_desc}")

    docstring = "\n".join(docstring)

    signature = inspect.Signature(parameters)

    async def dynamic_function(**kwargs):
        filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}

        if session is None:
            # will create a session one on the fly
            async with create_session(connection) as tool_session:
                await tool_session.initialize()
                call_tool_result = await cast(ClientSession, tool_session).call_tool(
                    mcp_tool.name, filtered_kwargs
                )
        else:
            call_tool_result = await session.call_tool(mcp_tool.name, filtered_kwargs)

        tool_message = _convert_call_tool_result(call_tool_result)
        return tool_message

    dynamic_function.__name__ = name
    dynamic_function.__doc__ = docstring
    dynamic_function.__signature__ = signature
    dynamic_function.__annotations__ = annotations

    return dynamic_function


def _json_schema_to_python_type(schema: dict) -> type:
    """Convert JSON schema type to Python type."""
    schema_type = schema.get("type", "string")

    type_mapping = {
        "string": str,
        "integer": int,
        "number": float,
        "boolean": bool,
        "array": list,
        "object": dict,
    }

    return type_mapping.get(schema_type, str)


def convert_mcp_tool_to_synalinks_tool(
    session: ClientSession | None,
    tool: MCPTool,
    *,
    connection: Connection | None = None,
    namespace: str | None = None,
) -> Tool:
    """Convert an MCP tool to a Synalinks tool.

    NOTE: tool can be executed only in a context of an active MCP client session.

    Args:
        session: MCP client session
        tool: MCP tool to convert
        connection: Optional connection config to use to create a new session
                    if a `session` is not provided
        namespace: Optional namespace to use for the tool name, if provided

    Returns:
        A Synalinks tool that wraps the MCP tool functionality
    """
    if session is None and connection is None:
        raise ValueError("Either a session or a connection config must be provided")

    function = _create_async_function_from_mcp_tool(
        tool, session, connection, namespace=namespace
    )
    return Tool(function)


async def load_mcp_tools(
    session: ClientSession | None,
    *,
    connection: Connection | None = None,
    namespace: str | None = None,
) -> list[Tool]:
    """Load all available MCP tools and convert them to Synalinks tools.

    Returns:
        A list of Synalinks tools with correct signature, annotations and schemas
    """
    if session is None and connection is None:
        raise ValueError("Either a session or a connection config must be provided")

    if session is None:
        # will create a session one on the fly
        async with create_session(connection) as tool_session:
            await tool_session.initialize()
            tools = await _list_all_tools(tool_session)
    else:
        tools = await _list_all_tools(session)

    converted_tools = [
        convert_mcp_tool_to_synalinks_tool(
            session, tool, connection=connection, namespace=namespace
        )
        for tool in tools
    ]
    return converted_tools



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import standardize_schema
from synalinks.src.modules import Input
from synalinks.src.modules import InputModule


class InputLayerTest(testing.TestCase):
    def test_input_basic(self):
        class Query(DataModel):
            query: str

        values = InputModule(input_data_model=SymbolicDataModel(data_model=Query))
        self.assertEqual(values.name, "input_module")
        self.assertEqual(values.get_schema(), standardize_schema(Query.get_schema()))

    def test_input(self):
        class Query(DataModel):
            query: str

        inputs = Input(data_model=Query)
        self.assertIsInstance(inputs, SymbolicDataModel)
        self.assertEqual(inputs.get_schema(), standardize_schema(Query.get_schema()))



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import getpass
import platform
import socket
import uuid
from functools import lru_cache

import mixpanel
import requests

from synalinks.src.hooks.hook import Hook
from synalinks.src.utils.async_utils import run_maybe_nested
from synalinks.src.version import version

_SYNALINKS_NAMESPACE = uuid.UUID("15f23928-68be-40c1-b69e-bc8b1fa71686")


_TELEMETRY = mixpanel.Mixpanel(
    "fbb605804abbb93efd661f5898c43cb5",
    consumer=mixpanel.Consumer(
        api_host="api-eu.mixpanel.com",
        retry_limit=2,
    ),
)


def device_id():
    """Generate a stable anonymous ID from hostname + username."""
    hostname = socket.gethostname()
    user = getpass.getuser()
    raw = f"{hostname}::{user}"
    return str(uuid.uuid5(_SYNALINKS_NAMESPACE, raw))


@lru_cache(maxsize=1)
def get_public_geo():
    resp = requests.get("https://ipapi.co/json/", timeout=3)
    resp.raise_for_status()
    data = resp.json()

    return {
        "city": data.get("city", "Unknown"),
        "region": data.get("region", "Unknown"),
        "country_code": data.get("country_code", "UN").upper(),
    }


class Telemetry(Hook):
    """
    Synalinks Telemetry Hook

    To disable the telemetry just use `synalinks.disable_telemetry()` at the
    beginning of your scripts

    ```python
    import synalinks

    synalinks.disable_telemetry()
    ```
    """

    async def _send(self, event, extra=None):
        try:
            geo = get_public_geo()

            props = {
                "module_class": self.module.__class__.__name__,
                "module_name": getattr(self.module, "name", "Unknown"),
                "module_description": getattr(self.module, "description", ""),
                "$os": platform.system(),
                "$city": geo["city"],
                "$region": geo["region"],
                "mp_country_code": geo["country_code"],
                "version": version(),
            }

            if extra:
                props.update(extra)

            _TELEMETRY.track(
                distinct_id=device_id(),
                event_name=event,
                properties=props,
            )

            # Identify this device/person once for Mixpanel People
            _TELEMETRY.people_set_once(device_id(), props)

            # Flush immediately
            if hasattr(_TELEMETRY, "flush"):
                _TELEMETRY.flush()

        except Exception:
            pass

    def on_call_begin(self, call_id, parent_call_id=None, inputs=None, kwargs=None):
        run_maybe_nested(self._send("call_begin_" + self.module.name))

    def on_call_end(self, call_id, parent_call_id=None, outputs=None, exception=None):
        if exception:
            run_maybe_nested(
                self._send("exception_" + self.module.name, {"exception": str(exception)})
            )
        else:
            run_maybe_nested(self._send("call_end_" + self.module.name))



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.language_models import LanguageModel
from synalinks.src.rewards.lm_as_judge import LMAsJudge


class LMAsJudgeTest(testing.TestCase):
    @patch("litellm.acompletion")
    async def test_lm_as_judge(self, mock_completion):
        class Query(DataModel):
            query: str = Field(description="The user query")

        class AnswerWithThinking(DataModel):
            thinking: str = Field(description="The step by step thinking process")
            answer: str = Field(description="The correct answer")

        language_model = LanguageModel(model="ollama/mistral")

        reward = LMAsJudge(language_model=language_model)

        inputs = Query(query="What is the French capital?")

        y_true = AnswerWithThinking(
            thinking="The French capital is Paris",
            answer="Paris",
        )

        y_pred = AnswerWithThinking(
            thinking="The French capital is well known",
            answer="Paris",
        )

        y_pred = inputs + y_pred

        expected_string = (
            """{"critique": "The answer is correct so we can attribute a high reward", """
            """"reward": 1.0}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        score = await reward(y_true=y_true, y_pred=y_pred)
        self.assertEqual(score, 1.0)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules import Input
from synalinks.src.modules.core.decision import Decision
from synalinks.src.programs import Program


class DecisionTest(testing.TestCase):
    @patch("litellm.acompletion")
    async def test_basic_decision(self, mock_completion):
        class Query(DataModel):
            query: str

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        expected_string = (
            """{"thinking": "The question ask for the capital of France, """
            """the answer is straitforward as it is well known that Paris is the"""
            """ capital", "choice": "easy"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        x0 = Input(data_model=Query)
        x1 = await Decision(
            question="What is the difficulty level of the above provided query?",
            labels=["easy", "difficult", "unkown"],
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
        )

        result = await program(Query(query="What is the French capital?"))

        self.assertEqual(result.get_json(), json.loads(expected_string))



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.modules.module import Module


@synalinks_export(
    [
        "synalinks.Xor",
        "synalinks.modules.Xor",
    ]
)
class Xor(Module):
    """Perform a logical Xor operation.

    It takes as input a list of data models,
    If more than two data models are not None, then it output None.
    otherwise it output the one that is not None.

    Table:

    | `x1`   | `x2`   | Logical Xor (`^`)|
    | ------ | ------ | ---------------- |
    | `x1`   | `x2`   | `None`           |
    | `x1`   | `None` | `x1`             |
    | `None` | `x2`   | `x2`             |
    | `None` | `None` | `None`           |

    Args:
        **kwargs (keyword arguments): Standard keyword arguments for the module.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def compute_output_spec(self, inputs, training=False):
        return inputs[0].clone()

    async def call(self, inputs, training=False):
        output = inputs[0]
        for i in range(1, len(inputs)):
            if inputs[i]:
                if not output:
                    output = inputs[i]
                else:
                    return None
        return output.clone(name=self.name)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import os
import tempfile
from datetime import datetime

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.backend import JsonDataModel
from synalinks.src.knowledge_bases import KnowledgeBase
from synalinks.src.modules import Input
from synalinks.src.modules.knowledge.stamp_knowledge import StampKnowledge
from synalinks.src.modules.knowledge.update_knowledge import UpdateKnowledge
from synalinks.src.programs import Program


class Document(DataModel):
    id: str = Field(description="The document id")
    text: str = Field(description="The content of the document")


class StampedDocument(DataModel):
    id: str = Field(description="The document id")
    text: str = Field(description="The content of the document")
    created_at: datetime = Field(description="The creation timestamp")


class StampKnowledgeTest(testing.TestCase):
    def setUp(self):
        super().setUp()
        self.temp_dir = tempfile.mkdtemp()
        self.db_path = os.path.join(self.temp_dir, "test.db")

    def tearDown(self):
        import shutil

        shutil.rmtree(self.temp_dir, ignore_errors=True)
        super().tearDown()

    async def test_stamp_knowledge_single(self):
        x0 = Input(data_model=Document)
        x1 = await StampKnowledge()(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="test_stamp_knowledge",
            description="test_stamp_knowledge",
        )

        input_doc = Document(id="doc1", text="test document")
        result = await program(input_doc)

        self.assertIsNotNone(result)
        self.assertTrue("created_at" in result)
        self.assertTrue("id" in result)
        self.assertTrue("text" in result)
        self.assertEqual(result.get("id"), "doc1")
        self.assertEqual(result.get("text"), "test document")

    async def test_stamp_knowledge_multiple(self):
        stamp_module = StampKnowledge()

        docs = [
            JsonDataModel(data_model=Document(id="doc1", text="first document")),
            JsonDataModel(data_model=Document(id="doc2", text="second document")),
        ]

        results = await stamp_module(docs)
        self.assertIsNotNone(results)
        self.assertEqual(len(results), 2)

        for result in results:
            self.assertTrue("created_at" in result)

    async def test_stamp_knowledge_none_input(self):
        stamp_module = StampKnowledge()
        result = await stamp_module(None)
        self.assertIsNone(result)

    def test_stamp_knowledge_serialization(self):
        stamp_module = StampKnowledge(
            name="test_stamp",
            description="Test stamp module",
        )

        config = stamp_module.get_config()
        cloned_module = StampKnowledge.from_config(config)

        self.assertEqual(cloned_module.name, "test_stamp")
        self.assertEqual(cloned_module.description, "Test stamp module")

    async def test_stamp_knowledge_contains_created_at_key(self):
        x0 = Input(data_model=Document)
        x1 = await StampKnowledge()(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="test_stamp_knowledge",
            description="test_stamp_knowledge",
        )

        input_doc = Document(id="doc1", text="test document")
        result = await program(input_doc)

        # Test the new string key contains functionality
        self.assertTrue("created_at" in result)
        self.assertFalse("nonexistent_key" in result)

    async def test_stamp_and_store_knowledge(self):
        knowledge_base = KnowledgeBase(
            uri=self.db_path,
            data_models=[StampedDocument],
        )

        x0 = Input(data_model=Document)
        x1 = await StampKnowledge()(x0)
        x2 = await UpdateKnowledge(knowledge_base=knowledge_base)(x1)

        program = Program(
            inputs=x0,
            outputs=x2,
            name="test_stamp_and_store",
            description="test_stamp_and_store",
        )

        input_doc = Document(id="doc1", text="test document")
        result = await program(input_doc)

        # Verify the stamped and stored result has all expected fields
        self.assertIsNotNone(result)
        self.assertTrue("created_at" in result)
        self.assertTrue("id" in result)
        self.assertTrue("text" in result)
        self.assertEqual(result.get("id"), "doc1")
        self.assertEqual(result.get("text"), "test document")
        self.assertIsNotNone(result.get("created_at"))

        # Retrieve stored data using get method
        symbolic_models = knowledge_base.get_symbolic_data_models()
        self.assertTrue(len(symbolic_models) > 0)

        retrieved = await knowledge_base.get("doc1", symbolic_models)
        self.assertIsNotNone(retrieved)
        self.assertEqual(retrieved.get_json()["id"], "doc1")
        self.assertEqual(retrieved.get_json()["text"], "test document")
        self.assertIn("created_at", retrieved.get_json())

    async def test_stamp_and_store_multiple(self):
        knowledge_base = KnowledgeBase(
            uri=self.db_path,
            data_models=[StampedDocument],
        )

        stamp_module = StampKnowledge()
        update_module = UpdateKnowledge(knowledge_base=knowledge_base)

        docs = [
            JsonDataModel(data_model=Document(id="doc1", text="first document")),
            JsonDataModel(data_model=Document(id="doc2", text="second document")),
        ]

        stamped = await stamp_module(docs)
        results = await update_module(stamped)

        # Verify the results have all expected fields
        self.assertIsNotNone(results)
        self.assertEqual(len(results), 2)

        for i, result in enumerate(results):
            self.assertTrue("created_at" in result)
            self.assertTrue("id" in result)
            self.assertTrue("text" in result)
            self.assertIsNotNone(result.get("created_at"))

        # Retrieve stored data using get method
        symbolic_models = knowledge_base.get_symbolic_data_models()
        self.assertTrue(len(symbolic_models) > 0)

        retrieved1 = await knowledge_base.get("doc1", symbolic_models)
        retrieved2 = await knowledge_base.get("doc2", symbolic_models)
        self.assertIsNotNone(retrieved1)
        self.assertIsNotNone(retrieved2)
        self.assertIn("created_at", retrieved1.get_json())
        self.assertIn("created_at", retrieved2.get_json())



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend.common.json_schema_utils import is_schema_equal
from synalinks.src.backend.common.json_schema_utils import standardize_schema
from synalinks.src.backend.pydantic.core import DataModel
from synalinks.src.backend.pydantic.core import is_meta_class


class CoreTest(testing.TestCase):
    def test_non_instanciated_data_model_schema(self):
        class TestDataModel(DataModel):
            foo: str

        expected_schema = {
            "additionalProperties": False,
            "properties": {"foo": {"title": "Foo", "type": "string"}},
            "required": ["foo"],
            "title": "TestDataModel",
            "type": "object",
        }

        self.assertEqual(TestDataModel.get_schema(), expected_schema)

    def test_instanciated_data_model_schema(self):
        class TestDataModel(DataModel):
            foo: str

        expected_schema = {
            "additionalProperties": False,
            "properties": {"foo": {"title": "Foo", "type": "string"}},
            "required": ["foo"],
            "title": "TestDataModel",
            "type": "object",
        }

        self.assertEqual(TestDataModel(foo="bar").get_schema(), expected_schema)

    def test_instanciated_data_model_json(self):
        class TestDataModel(DataModel):
            foo: str

        expected_json = {"foo": "bar"}

        self.assertEqual(TestDataModel(foo="bar").get_json(), expected_json)

    def test_concatenate_meta_data_model(self):
        class Foo(DataModel):
            foo: str

        class Bar(DataModel):
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        x = Foo + Bar

        schema = x.get_schema()
        expected_schema = standardize_schema(Result.get_schema())
        self.assertTrue(is_schema_equal(schema, expected_schema))

    def test_and_meta_data_model(self):
        class Foo(DataModel):
            foo: str

        class Bar(DataModel):
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        x = Foo & Bar

        schema = x.get_schema()
        expected_schema = standardize_schema(Result.get_schema())
        self.assertTrue(is_schema_equal(schema, expected_schema))

    def test_or_meta_data_model(self):
        class Foo(DataModel):
            foo: str

        class Bar(DataModel):
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        x = Foo | Bar

        schema = x.get_schema()
        expected_schema = standardize_schema(Foo.get_schema())
        self.assertEqual(schema, expected_schema)

    def test_xor_meta_data_model(self):
        class Foo(DataModel):
            foo: str

        class Bar(DataModel):
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        x = Foo ^ Bar

        schema = x.get_schema()
        expected_schema = standardize_schema(Foo.get_schema())
        self.assertEqual(schema, expected_schema)

    def test_not_meta_data_model(self):
        class Foo(DataModel):
            foo: str

        x = ~Foo

        schema = x.get_schema()
        expected_schema = standardize_schema(Foo.get_schema())
        self.assertEqual(schema, expected_schema)

    def test_is_meta_class(self):
        class Query(DataModel):
            query: str

        self.assertTrue(is_meta_class(Query))

    def test_is_not_meta_class(self):
        class Query(DataModel):
            query: str

        self.assertFalse(is_meta_class(Query(query="What is the French capital?")))

    def test_contains_meta_class(self):
        class Foo(DataModel):
            foo: str

        class FooBar(DataModel):
            foo: str
            bar: str

        class Bar(DataModel):
            bar: str

        self.assertTrue(Foo in FooBar)
        self.assertFalse(Bar in Foo)

    def test_contains_string_key_meta_class(self):
        class FooBar(DataModel):
            foo: str
            bar: str

        self.assertTrue("foo" in FooBar)
        self.assertTrue("bar" in FooBar)
        self.assertFalse("baz" in FooBar)

    def test_contains_data_model_instance(self):
        class Foo(DataModel):
            foo: str

        class FooBar(DataModel):
            foo: str
            bar: str

        class Bar(DataModel):
            bar: str

        foo_instance = Foo(foo="a")
        foobar_instance = FooBar(foo="a", bar="b")
        bar_instance = Bar(bar="c")

        self.assertTrue(foo_instance in foobar_instance)
        self.assertFalse(bar_instance in foo_instance)

    def test_contains_string_key_data_model_instance(self):
        class FooBar(DataModel):
            foo: str
            bar: str

        foobar_instance = FooBar(foo="a", bar="b")

        self.assertTrue("foo" in foobar_instance)
        self.assertTrue("bar" in foobar_instance)
        self.assertFalse("baz" in foobar_instance)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import ChatMessage
from synalinks.src.backend import ChatMessages
from synalinks.src.backend import is_chat_messages
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules.agents.function_calling_agent import FunctionCallingAgent
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.core.tool import Tool
from synalinks.src.programs import Program
from synalinks.src.saving.object_registration import register_synalinks_serializable


@register_synalinks_serializable()
async def calculate(expression: str):
    """Calculate the result of a mathematical expression.

    Args:
        expression (str): The mathematical expression to calculate, such as
            '2 + 2'. The expression can contain numbers, operators (+, -, *, /),
            parentheses, and spaces.
    """
    if not all(char in "0123456789+-*/(). " for char in expression):
        return {
            "result": None,
            "log": "Error: invalid characters in expression",
        }
    try:
        # Evaluate the mathematical expression safely
        result = round(float(eval(expression, {"__builtins__": None}, {})), 2)
        return {
            "result": result,
            "log": "Successfully executed",
        }
    except Exception as e:
        return {
            "result": None,
            "log": f"Error: {e}",
        }


@register_synalinks_serializable()
async def thinking(thinking: str):
    """Think about something.

    Args:
        thinking (str): Your step by step thinking.
    """
    return {
        "thinking": thinking,
    }


@register_synalinks_serializable()
async def get_weather(location: str):
    """Get weather information for a location.
    Args:
        location (str): The location to get weather for.
    """
    # Mock weather data
    weather_data = {
        "New York": {"temp": 22, "condition": "Sunny"},
        "London": {"temp": 15, "condition": "Cloudy"},
        "Tokyo": {"temp": 28, "condition": "Rainy"},
    }

    if location in weather_data:
        return {
            "location": location,
            "temperature": weather_data[location]["temp"],
            "condition": weather_data[location]["condition"],
            "success": True,
        }
    else:
        return {"location": location, "error": "Location not found", "success": False}


@register_synalinks_serializable()
async def failing_tool(should_fail: bool = True):
    """A tool that intentionally fails for testing error handling.
    Args:
        should_fail (bool): Whether the tool should fail.
    """
    if should_fail:
        raise ValueError("This tool was designed to fail")
    return {"status": "success"}


class FunctionCallingAgentTest(testing.TestCase):
    async def test_agent_instantiation(self):
        """Test basic agent instantiation."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
            Tool(thinking),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
        )(inputs)
        program = Program(
            inputs=inputs,
            outputs=outputs,
            name="function_calling_agent_test",
        )
        self.assertIsNotNone(program)

    async def test_agent_temperature_parameter(self):
        """Test that temperature is correctly passed to generators."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [Tool(calculate)]

        agent = FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            temperature=0.7,
            name="temp_test",
        )

        # Verify temperature is stored
        self.assertEqual(agent.temperature, 0.7)
        # Verify temperature is passed to tool_calls_generator
        self.assertEqual(agent.tool_calls_generator.temperature, 0.7)

    async def test_agent_reasoning_effort_parameter(self):
        """Test that reasoning_effort is correctly passed to generators."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [Tool(calculate)]

        agent = FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            reasoning_effort="low",
            name="reasoning_test",
        )

        # Verify reasoning_effort is stored
        self.assertEqual(agent.reasoning_effort, "low")
        # Verify reasoning_effort is passed to tool_calls_generator
        self.assertEqual(agent.tool_calls_generator.reasoning_effort, "low")

    async def test_agent_default_reasoning_effort_is_none(self):
        """Test that default reasoning_effort is None (not 'low' like ChainOfThought)."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [Tool(calculate)]

        agent = FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            name="default_test",
        )

        # FunctionCallingAgent should default to None (no reasoning)
        # But its ChainOfThought generators will default to "low"
        self.assertIsNone(agent.reasoning_effort)
        # The ChainOfThought inside defaults None to "low"
        self.assertEqual(agent.tool_calls_generator.reasoning_effort, "low")

    @patch("litellm.acompletion")
    async def test_autonomous_mode_simple_calculation(self, mock_completion):
        """Test autonomous mode with a simple calculation."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
            Tool(thinking),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=True,
            max_iterations=3,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="autonomous_calculation_test",
        )

        tool_calls = {
            "thinking": (
                "Perform simple arithmetic operation by adding the numbers "
                "given in the input."
            ),
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "152648 + 485",
                }
            ],
        }

        tool_calls_1 = {
            "thinking": (
                "The user has asked for a simple arithmetic operation, "
                "specifically adding 152648 and 485. I have already performed "
                "the calculation using the 'calculate' tool and obtained the "
                "result as 153133."
            ),
            "tool_calls": [],
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls)}}]},
            {"choices": [{"message": {"content": json.dumps(tool_calls_1)}}]},
        ]

        mock_completion.side_effect = mock_responses

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content="How much is 152648 + 485?",
                ),
            ]
        )
        result = await agent(input_messages)

        print("Result:")
        print(result.prettify_json())

        # Verify result structure
        self.assertIsNotNone(result)
        messages = result.get("messages", [])
        self.assertGreater(len(messages), 0)
        self.assertTrue(is_chat_messages(result))

    @patch("litellm.acompletion")
    async def test_autonomous_mode_complex_calculation(self, mock_completion):
        """Test autonomous mode with a more complex calculation."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
            Tool(thinking),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=True,
            max_iterations=5,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="complex_calculation_test",
        )

        tool_calls = {
            "thinking": (
                "First, I will perform the arithmetic operation as instructed. "
                "Let's calculate (150 + 250) * 2 / 4. The order of operations "
                "is follow BIDMAS/BODMAS which means Brackets, Orders or "
                "Powers, Division and Multiplication, Addition and "
                "Subtraction. So, first I will add 150 and 250, then multiply "
                "the result by 2, divide it by 4 and finally add 100."
            ),
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "(150 + 250) * 2 / 4 + 100",
                }
            ],
        }

        tool_calls_1 = {
            "thinking": (
                "The user provided a mathematical expression to calculate. I "
                "performed the operation (150 + 250) * 2 / 4 and then added "
                "100 to the result. Now, the result is 300."
            ),
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "(150 + 250) * 2 / 4 + 100",
                }
            ],
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls)}}]},
            {"choices": [{"message": {"content": json.dumps(tool_calls_1)}}]},
        ]

        mock_completion.side_effect = mock_responses

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content=(
                        "Calculate (150 + 250) * 2 / 4 and then add 100 to the result"
                    ),
                ),
            ]
        )
        result = await agent(input_messages)
        print("Result:")
        print(result.prettify_json())

    @patch("litellm.acompletion")
    async def test_autonomous_mode_returns_chat_messages_without_schema(
        self, mock_completion
    ):
        """Test that autonomous mode returns ChatMessages when no schema is provided."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=True,
            max_iterations=3,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="autonomous_no_schema_test",
        )

        tool_calls = {
            "thinking": "I need to calculate 10 + 20.",
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "10 + 20",
                }
            ],
        }

        tool_calls_1 = {
            "thinking": "The calculation is complete. The result is 30.",
            "tool_calls": [],
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls)}}]},
            {"choices": [{"message": {"content": json.dumps(tool_calls_1)}}]},
        ]

        mock_completion.side_effect = mock_responses

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content="How much is 10 + 20?",
                ),
            ]
        )
        result = await agent(input_messages)

        # Verify result is ChatMessages
        self.assertIsNotNone(result)
        self.assertTrue(is_chat_messages(result))

        # Verify messages include tool calls and tool results
        messages = result.get("messages", [])
        self.assertGreater(len(messages), 1)

        # Check that we have user, assistant with tool_calls, tool, and final assistant
        roles = [msg.get("role") for msg in messages]
        self.assertIn("user", roles)
        self.assertIn("assistant", roles)
        self.assertIn("tool", roles)

    @patch("litellm.acompletion")
    async def test_non_autonomous_mode_returns_chat_messages(self, mock_completion):
        """Test that non-autonomous mode returns ChatMessages."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=False,
            return_inputs_with_trajectory=False,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="non_autonomous_test",
        )

        tool_calls = {
            "thinking": "I need to calculate 5 * 5.",
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "5 * 5",
                }
            ],
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls)}}]},
        ]

        mock_completion.side_effect = mock_responses

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content="How much is 5 * 5?",
                ),
            ]
        )
        result = await agent(input_messages)

        # Verify result is ChatMessages (not ChatMessage)
        self.assertIsNotNone(result)
        self.assertTrue(is_chat_messages(result))

        # Verify messages structure - only assistant message since no prior tool calls
        messages = result.get("messages", [])
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].get("role"), "assistant")

    @patch("litellm.acompletion")
    async def test_non_autonomous_mode_returns_tool_and_assistant_messages(
        self, mock_completion
    ):
        """Test non-autonomous mode returns tool messages and assistant message."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=False,
            return_inputs_with_trajectory=False,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="non_autonomous_tool_messages_test",
        )

        # Second LLM response after tool execution (no more tool calls)
        tool_calls_response = {
            "thinking": "The calculation result is 25. Task complete.",
            "tool_calls": [],
        }

        # Final generator response (ChatMessage format since no schema)
        final_response = {
            "role": "assistant",
            "content": "The result of 5 * 5 is 25.",
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls_response)}}]},
            {"choices": [{"message": {"content": json.dumps(final_response)}}]},
        ]

        mock_completion.side_effect = mock_responses

        # Input includes a previous assistant message with tool_calls
        # This simulates continuing the conversation after the user approved tool calls
        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content="How much is 5 * 5?",
                ),
                ChatMessage(
                    role="assistant",
                    content="I need to calculate 5 * 5.",
                    tool_calls=[
                        {
                            "id": "test-tool-call-id",
                            "name": "calculate",
                            "arguments": {"expression": "5 * 5"},
                        }
                    ],
                ),
            ]
        )
        result = await agent(input_messages)

        # Verify result contains the new messages only (tool result + final assistant)
        # With return_inputs_with_trajectory=False, only new messages are returned
        self.assertIsNotNone(result)
        self.assertTrue(is_chat_messages(result))

        # The result should contain only the new messages (tool result + final assistant)
        messages = result.get("messages", [])
        self.assertEqual(len(messages), 2)

        # First message should be the tool result
        self.assertEqual(messages[0].get("role"), "tool")
        self.assertEqual(messages[0].get("tool_call_id"), "test-tool-call-id")

        # Second message should be the final assistant response
        self.assertEqual(messages[1].get("role"), "assistant")

    @patch("litellm.acompletion")
    async def test_non_autonomous_mode_with_trajectory(self, mock_completion):
        """Test non-autonomous mode with return_inputs_with_trajectory=True."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=False,
            return_inputs_with_trajectory=True,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="non_autonomous_trajectory_test",
        )

        tool_calls = {
            "thinking": "I need to calculate 3 + 3.",
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "3 + 3",
                }
            ],
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls)}}]},
        ]

        mock_completion.side_effect = mock_responses

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content="How much is 3 + 3?",
                ),
            ]
        )
        result = await agent(input_messages)

        # Verify result is ChatMessages
        self.assertIsNotNone(result)
        self.assertTrue(is_chat_messages(result))

        # Verify full trajectory is returned (user message + assistant message)
        messages = result.get("messages", [])
        self.assertGreaterEqual(len(messages), 2)
        self.assertEqual(messages[0].get("role"), "user")
        self.assertEqual(messages[-1].get("role"), "assistant")

    @patch("litellm.acompletion")
    async def test_interactive_mode_single_step(self, mock_completion):
        """Test interactive mode with single step execution."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
            Tool(thinking),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=False,
            return_inputs_with_trajectory=True,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="interactive_single_step_test",
        )

        tool_calls = {
            "thinking": "I need to calculate 152648 + 485.",
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "152648 + 485",
                }
            ],
        }

        mock_completion.return_value = {
            "choices": [{"message": {"content": json.dumps(tool_calls)}}]
        }

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content="How much is 152648 + 485?",
                )
            ]
        )
        result = await agent(input_messages)

        # Verify result structure
        self.assertIsNotNone(result)
        self.assertTrue(is_chat_messages(result))
        messages = result.get("messages", [])
        self.assertGreaterEqual(len(messages), 2)
        self.assertEqual(messages[0].get("role"), "user")
        self.assertEqual(messages[-1].get("role"), "assistant")

    @patch("litellm.acompletion")
    async def test_interactive_mode_multi_step(self, mock_completion):
        """Test interactive mode with multiple steps simulation."""
        language_model = LanguageModel(model="ollama/mistral")
        tools = [
            Tool(calculate),
            Tool(thinking),
        ]
        inputs = Input(data_model=ChatMessages)
        outputs = await FunctionCallingAgent(
            language_model=language_model,
            tools=tools,
            autonomous=False,
            return_inputs_with_trajectory=True,
        )(inputs)
        agent = Program(
            inputs=inputs,
            outputs=outputs,
            name="interactive_multi_step_test",
        )

        # First step: calculate 100 + 200
        tool_calls_step1 = {
            "thinking": "First, I need to calculate 100 + 200.",
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "100 + 200",
                }
            ],
        }

        # Second step: multiply result by 3
        tool_calls_step2 = {
            "thinking": "Now I need to multiply 300 by 3.",
            "tool_calls": [
                {
                    "tool_name": "calculate",
                    "expression": "300 * 3",
                }
            ],
        }

        # Final step: no more tool calls
        tool_calls_step3 = {
            "thinking": "Calculation complete. 100 + 200 = 300, then 300 * 3 = 900.",
            "tool_calls": [],
        }

        # Final generator response (ChatMessage format since no schema)
        final_response = {
            "role": "assistant",
            "content": "The calculation is complete."
            " 100 + 200 = 300, then 300 * 3 = 900.",
        }

        mock_responses = [
            {"choices": [{"message": {"content": json.dumps(tool_calls_step1)}}]},
            {"choices": [{"message": {"content": json.dumps(tool_calls_step2)}}]},
            {"choices": [{"message": {"content": json.dumps(tool_calls_step3)}}]},
            {"choices": [{"message": {"content": json.dumps(final_response)}}]},
        ]
        mock_completion.side_effect = mock_responses

        input_messages = ChatMessages(
            messages=[
                ChatMessage(
                    role="user",
                    content="I need to calculate 100 + 200, then multiply by 3",
                )
            ]
        )

        # Simulate multiple interaction steps
        max_steps = 3
        for step in range(max_steps):
            result = await agent(input_messages)

            # Verify result is ChatMessages
            self.assertIsNotNone(result)
            self.assertTrue(is_chat_messages(result))

            # Get the latest assistant message
            messages = result.get("messages", [])
            if messages:
                last_message = messages[-1]
                if last_message.get("role") == "assistant":
                    tool_calls = last_message.get("tool_calls", [])
                    if not tool_calls:
                        break
                    # Continue with the result as new input
                    input_messages = result
                else:
                    break
            else:
                break

        # Verify we completed all steps
        self.assertEqual(step, 2)  # 0, 1, 2 = 3 steps



# Modified from: keras/src/callbacks/callback_test.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.callbacks import Callback
from synalinks.src.optimizers.random_few_shot import RandomFewShot
from synalinks.src.programs import Program
from synalinks.src.rewards import ExactMatch


class CallbackTest(testing.TestCase):
    def test_model_state_is_current_on_epoch_end(self):
        class Query(DataModel):
            query: str

        class Answer(DataModel):
            answer: str

        class Iterations(DataModel):
            count: int = 0

        class TestProgram(Program):
            def __init__(self):
                super().__init__()
                self.iterations = self.add_variable(
                    data_model=Iterations,
                )

            def call(self, inputs):
                self.iterations.json["count"] += 1
                return inputs

        class CBK(Callback):
            def on_batch_end(self, batch, logs):
                assert self.iterations.json["count"] == batch + 1

        model = TestProgram()
        model.compile(optimizer=RandomFewShot(), reward=ExactMatch())
        x = [Query(query="test 1"), Query(query="test 2")]
        y = [Answer(answer="answer 1"), Answer(answer="answer 2")]
        model.fit(x, y, callbacks=[CBK()], batch_size=2)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import List

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import is_schema_equal
from synalinks.src.backend import standardize_schema
from synalinks.src.modules import Input
from synalinks.src.ops.json import concat
from synalinks.src.ops.json import factorize
from synalinks.src.ops.json import in_mask
from synalinks.src.ops.json import logical_and
from synalinks.src.ops.json import logical_or
from synalinks.src.ops.json import logical_xor
from synalinks.src.ops.json import out_mask
from synalinks.src.programs import Program
from synalinks.src.utils.nlp_utils import remove_numerical_suffix


class SchemaEqualTest(testing.TestCase):
    def test_is_schema_equal(self):
        class Test(DataModel):
            foo: str
            foo_1: str

        x = SymbolicDataModel(data_model=Test)
        y = SymbolicDataModel(data_model=Test)
        self.assertEqual(x.get_schema(), standardize_schema(Test.get_schema()))
        self.assertEqual(y.get_schema(), standardize_schema(Test.get_schema()))
        self.assertTrue(is_schema_equal(x.get_schema(), y.get_schema()))

    def test_schema_not_equal(self):
        class Test1(DataModel):
            foo: str

        class Test2(DataModel):
            foo: str
            foo_1: str

        x = SymbolicDataModel(data_model=Test1)
        y = SymbolicDataModel(data_model=Test2)
        self.assertEqual(x.get_schema(), standardize_schema(Test1.get_schema()))
        self.assertEqual(y.get_schema(), standardize_schema(Test2.get_schema()))
        self.assertFalse(is_schema_equal(x.get_schema(), y.get_schema()))

    def test_not_equal_but_same_names(self):
        class Test1(DataModel):
            foo: str

        class Test2(DataModel):
            foo: int

        x = SymbolicDataModel(data_model=Test1)
        y = SymbolicDataModel(data_model=Test2)
        self.assertEqual(x.get_schema(), standardize_schema(Test1.get_schema()))
        self.assertEqual(y.get_schema(), standardize_schema(Test2.get_schema()))
        self.assertFalse(is_schema_equal(x.get_schema(), y.get_schema()))


class ConcatenateTest(testing.TestCase):
    async def test_concat_schema_same_property(self):
        class Test1(DataModel):
            foo: str

        class Test2(DataModel):
            foo: str

        class Result(DataModel):
            foo: str
            foo_1: str

        x = SymbolicDataModel(data_model=Test1)
        y = SymbolicDataModel(data_model=Test2)
        expected = SymbolicDataModel(data_model=Result)

        result = await concat(x, y)
        self.assertTrue(is_schema_equal(result.get_schema(), expected.get_schema()))

    async def test_concat_schema_different_property(self):
        class Test1(DataModel):
            foo: str

        class Test2(DataModel):
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        x = SymbolicDataModel(data_model=Test1)
        y = SymbolicDataModel(data_model=Test2)
        expected = SymbolicDataModel(data_model=Result)
        result = await concat(x, y)
        self.assertTrue(is_schema_equal(result.get_schema(), expected.get_schema()))

    async def test_concat_serialization(self):
        class Test1(DataModel):
            foo: str

        class Test2(DataModel):
            bar: str

        i0 = Input(data_model=Test1)
        i1 = Input(data_model=Test2)
        x0 = await concat(i0, i1)

        program = Program(
            inputs=[i0, i1],
            outputs=x0,
        )

        config = program.get_config()
        new_program = Program.from_config(config)
        for var1 in program.variables:
            for var2 in new_program.variables:
                if remove_numerical_suffix(var2.path) == var1.path:
                    self.assertEqual(var2.get_json(), var1.get_json())


class FactorizeTest(testing.TestCase):
    async def test_factorize_schema_same_property(self):
        class Test(DataModel):
            foo: str
            foo_1: str

        class Result(DataModel):
            foos: List[str]

        x = SymbolicDataModel(data_model=Test)
        expected = SymbolicDataModel(data_model=Result)
        result = await factorize(x)
        self.assertTrue(is_schema_equal(result.get_schema(), expected.get_schema()))

    async def test_factorize_schema_properties_multiple_times(self):
        class Test(DataModel):
            foos: List[str]
            foo: str

        class Result(DataModel):
            foos: List[str]

        x = SymbolicDataModel(data_model=Test)
        expected = SymbolicDataModel(data_model=Result)
        result = await factorize(x)
        self.assertTrue(is_schema_equal(result.get_schema(), expected.get_schema()))

    async def test_factorize_schema_different_property(self):
        class Test(DataModel):
            foo: str
            bar: str

        class Result(DataModel):
            foo: str
            bar: str

        x = SymbolicDataModel(data_model=Test)
        expected = SymbolicDataModel(data_model=Result)
        result = await factorize(x)
        self.assertTrue(is_schema_equal(result.get_schema(), expected.get_schema()))

    async def test_factorize_serialization(self):
        class Test(DataModel):
            foos: List[str]
            foo: str

        i0 = Input(data_model=Test)
        x0 = await factorize(i0)

        program = Program(
            inputs=i0,
            outputs=x0,
        )

        config = program.get_config()
        new_program = Program.from_config(config)
        for var1 in program.variables:
            for var2 in new_program.variables:
                if remove_numerical_suffix(var2.path) == var1.path:
                    self.assertEqual(var2.get_json(), var1.get_json())


class LogicalAndTest(testing.TestCase):
    async def test_and_table(self):
        class Answer(DataModel):
            answer: str

        class Result(DataModel):
            answer: str
            answer_1: str

        i0 = Input(data_model=Answer)
        i1 = Input(data_model=Answer)

        x0 = await logical_and(i0, i1)

        program = Program(
            inputs=[i0, i1],
            outputs=x0,
        )

        result = await program([Answer(answer="Paris"), Answer(answer="Toulouse")])
        self.assertEqual(
            result.get_json(),
            Result(
                answer="Paris",
                answer_1="Toulouse",
            ).get_json(),
        )

        result = await program([Answer(answer="Paris"), None])
        self.assertEqual(result, None)

        result = await program([None, Answer(answer="Toulouse")])
        self.assertEqual(result, None)

        result = await program([None, None])
        self.assertEqual(result, None)


class LogicalOrTest(testing.TestCase):
    async def test_or_table(self):
        class Answer(DataModel):
            answer: str

        class Result(DataModel):
            answer: str
            answer_1: str

        i0 = Input(data_model=Answer)
        i1 = Input(data_model=Answer)

        x0 = await logical_or(i0, i1)

        program = Program(
            inputs=[i0, i1],
            outputs=x0,
        )

        result = await program([Answer(answer="Paris"), Answer(answer="Toulouse")])
        self.assertEqual(
            result.get_json(),
            Result(
                answer="Paris",
                answer_1="Toulouse",
            ).get_json(),
        )

        result = await program([Answer(answer="Paris"), None])
        self.assertEqual(result.get_json(), Answer(answer="Paris").get_json())

        result = await program([None, Answer(answer="Toulouse")])
        self.assertEqual(result.get_json(), Answer(answer="Toulouse").get_json())

        result = await program([None, None])
        self.assertEqual(result, None)


class LogicalXorTest(testing.TestCase):
    async def test_xor_table(self):
        class Answer(DataModel):
            answer: str

        class Result(DataModel):
            answer: str
            answer_1: str

        i0 = Input(data_model=Answer)
        i1 = Input(data_model=Answer)

        x0 = await logical_xor(i0, i1)

        program = Program(
            inputs=[i0, i1],
            outputs=x0,
        )

        result = await program([Answer(answer="Paris"), Answer(answer="Toulouse")])
        self.assertEqual(result, None)

        result = await program([Answer(answer="Paris"), None])
        self.assertEqual(result.get_json(), Answer(answer="Paris").get_json())

        result = await program([None, Answer(answer="Toulouse")])
        self.assertEqual(result.get_json(), Answer(answer="Toulouse").get_json())

        result = await program([None, None])
        self.assertEqual(result, None)


class OutMaskTest(testing.TestCase):
    async def test_out_mask_serialization(self):
        class Test(DataModel):
            foos: List[str]
            foo: str
            bar: str

        i0 = Input(data_model=Test)
        x0 = await out_mask(i0, mask=["foo"])

        program = Program(
            inputs=i0,
            outputs=x0,
        )

        config = program.get_config()
        new_program = Program.from_config(config)
        for var1 in program.variables:
            for var2 in new_program.variables:
                if remove_numerical_suffix(var2.path) == var1.path:
                    self.assertEqual(var2.get_json(), var1.get_json())


class InMaskTest(testing.TestCase):
    async def test_in_mask_serialization(self):
        class Test(DataModel):
            foos: List[str]
            foo: str
            bar: str

        i0 = Input(data_model=Test)
        x0 = await in_mask(i0, mask=["bar"])

        program = Program(
            inputs=i0,
            outputs=x0,
        )

        config = program.get_config()
        new_program = Program.from_config(config)
        for var1 in program.variables:
            for var2 in new_program.variables:
                if remove_numerical_suffix(var2.path) == var1.path:
                    self.assertEqual(var2.get_json(), var1.get_json())



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import saving
from synalinks.src import testing
from synalinks.src.utils.tool_utils import Tool


@saving.object_registration.register_synalinks_serializable()
async def calculate(expression: str):
    """Calculate the result of a mathematical expression.

    Args:
        expression (str): The mathematical expression to calculate, such as
            '2 + 2'. The expression can contain numbers, operators (+, -, *, /),
            parentheses, and spaces.
    """
    if not all(char in "0123456789+-*/(). " for char in expression):
        return {
            "result": None,
            "log": "Error: invalid characters in expression",
        }
    try:
        # Evaluate the mathematical expression safely
        result = round(float(eval(expression, {"__builtins__": None}, {})), 2)
        return {
            "result": result,
            "log": "Successfully executed",
        }
    except Exception as e:
        return {
            "result": None,
            "log": f"Error: {e}",
        }


class ToolUtilsTest(testing.TestCase):
    def test_basic_tool(self):
        _ = Tool(calculate)

    async def test_tool_serialization(self):
        tool = Tool(calculate)
        tool_config = tool.get_config()
        new_tool = Tool.from_config(tool_config)

        tool_call = await new_tool("2+2")
        result = tool_call.get("result")

        self.assertTrue(result == 4)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import numpy as np

from synalinks.src.api_export import synalinks_export
from synalinks.src.optimizers.optimizer import Optimizer


@synalinks_export("synalinks.optimizers.GreedyOptimizer")
class GreedyOptimizer(Optimizer):
    """Abstract base class for greedy sampling-based optimizers.

    Greedy optimizers focus on sampling from best predictions.
    They use configurable sampling strategies to select high-reward
    predictions as few-shot examples.

    Subclasses must implement:
        - `propose_new_candidates()`: Generate new candidates using sampling

    Args:
        nb_min_examples (int): The min number of examples for few-shot
            learning (Default to 1).
        nb_max_examples (int): The max number of examples for few-shot
            learning (Default to 3).
        sampling (str): The method to sample predictions between
            ['random', 'best', 'softmax']. (Default to 'softmax').
        sampling_temperature (float): The temperature for softmax sampling.
            Used only when `sampling='softmax'`. Lower values concentrate
            sampling on high-reward predictions, higher values make sampling
            more uniform (Default 0.3).
        population_size (int): The maximum number of best candidates to keep
            during the optimization process.
        name (str): Optional name for the optimizer instance.
        description (str): Optional description of the optimizer instance.
    """

    def __init__(
        self,
        nb_min_examples=1,
        nb_max_examples=3,
        sampling="softmax",
        sampling_temperature=0.3,
        population_size=10,
        name=None,
        description=None,
        **kwargs,
    ):
        super().__init__(
            population_size=population_size,
            name=name,
            description=description,
            **kwargs,
        )
        self.nb_min_examples = nb_min_examples
        self.nb_max_examples = nb_max_examples

        samplings = ["random", "best", "softmax"]
        if sampling not in samplings:
            raise ValueError(f"Parameter `sampling` should be between {samplings}")
        self.sampling = sampling
        self.sampling_temperature = sampling_temperature

    async def sample_best_predictions(
        self,
        trainable_variable,
    ):
        """Sample predictions based on the configured sampling strategy.

        Samples between `nb_min_examples` and `nb_max_examples` predictions
        from the trainable variable using the configured sampling method:
        - 'random': uniformly random selection
        - 'best': select highest reward predictions
        - 'softmax': softmax sampling biased toward higher rewards

        Args:
            trainable_variable: The trainable variable containing predictions.

        Returns:
            List of selected predictions.
        """
        predictions = trainable_variable.get("predictions")
        nb_examples = np.random.randint(self.nb_min_examples, self.nb_max_examples + 1)
        selected_predictions = []

        if nb_examples != 0:
            if len(predictions) > nb_examples:
                if self.sampling == "random":
                    selected_predictions = np.random.choice(
                        predictions,
                        size=nb_examples,
                        replace=False,
                    ).tolist()
                elif self.sampling == "best":
                    sorted_predictions = sorted(
                        predictions,
                        key=lambda x: x.get("reward", 0),
                        reverse=True,
                    )
                    selected_predictions = sorted_predictions[:nb_examples]
                elif self.sampling == "softmax":
                    rewards = np.array([pred.get("reward", 0) for pred in predictions])
                    scaled_rewards = rewards / self.sampling_temperature
                    exp_rewards = np.exp(scaled_rewards - np.max(scaled_rewards))
                    probabilities = exp_rewards / np.sum(exp_rewards)
                    selected_predictions = np.random.choice(
                        predictions,
                        size=nb_examples,
                        replace=False,
                        p=probabilities,
                    ).tolist()
            else:
                selected_predictions = predictions

        return selected_predictions

    def get_config(self):
        base_config = super().get_config()
        base_config.update(
            {
                "nb_min_examples": self.nb_min_examples,
                "nb_max_examples": self.nb_max_examples,
                "sampling": self.sampling,
                "sampling_temperature": self.sampling_temperature,
            }
        )
        return base_config



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from datetime import datetime

from synalinks.src import ops
from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import Stamp
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.modules.module import Module
from synalinks.src.utils.async_utils import run_maybe_nested


@synalinks_export(
    [
        "synalinks.modules.StampKnowledge",
        "synalinks.StampKnowledge",
    ]
)
class StampKnowledge(Module):
    """Stamps data models with a creation timestamp.

    This module adds a `created_at` field containing the current datetime
    to each input data model using a logical AND operation.

    Args:
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        name=None,
        description=None,
        trainable=False,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )

    async def _stamp(self, data_model):
        timestamp = JsonDataModel(
            json={"created_at": datetime.now().isoformat()},
            schema=Stamp.get_schema(),
            name="timestamp_" + data_model.name,
        )
        return await ops.logical_and(
            data_model,
            timestamp,
            name="stamped_" + data_model.name,
        )

    async def call(self, inputs):
        if not inputs:
            return None
        return tree.map_structure(
            lambda x: run_maybe_nested(self._stamp(x)),
            inputs,
        )

    async def compute_output_spec(self, inputs):
        def _stamp_spec(x):
            timestamp = SymbolicDataModel(
                schema=Stamp.get_schema(),
                name="timestamp_" + x.name,
            )
            return run_maybe_nested(
                ops.logical_and(
                    x,
                    timestamp,
                    name="stamped_" + x.name,
                )
            )

        return tree.map_structure(_stamp_spec, inputs)

    def get_config(self):
        return {
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import MagicMock

from synalinks.src import testing
from synalinks.src.backend import JsonDataModel
from synalinks.src.optimizers.evolutionary_optimizer import EvolutionaryOptimizer
from synalinks.src.optimizers.omega import OMEGA
from synalinks.src.optimizers.optimizer import Optimizer


class EvolutionaryOptimizerTest(testing.TestCase):
    def test_inheritance(self):
        """Test that EvolutionaryOptimizer inherits from Optimizer."""
        self.assertTrue(issubclass(EvolutionaryOptimizer, Optimizer))
        self.assertTrue(issubclass(OMEGA, EvolutionaryOptimizer))

    def test_init_default_parameters(self):
        """Test initialization with default parameters."""
        optimizer = EvolutionaryOptimizer()

        self.assertIsNone(optimizer.language_model)
        self.assertEqual(optimizer.mutation_temperature, 0.3)
        self.assertEqual(optimizer.crossover_temperature, 0.3)
        self.assertEqual(optimizer.selection, "softmax")
        self.assertEqual(optimizer.selection_temperature, 0.3)
        self.assertEqual(optimizer.merging_rate, 0.02)
        self.assertEqual(optimizer.population_size, 10)

    def test_init_custom_parameters(self):
        """Test initialization with custom parameters."""
        mock_lm = MagicMock()

        optimizer = EvolutionaryOptimizer(
            language_model=mock_lm,
            mutation_temperature=0.5,
            crossover_temperature=0.7,
            selection="best",
            selection_temperature=0.2,
            merging_rate=0.1,
            population_size=20,
            name="test_evo_optimizer",
            description="Test evolutionary optimizer",
        )

        self.assertEqual(optimizer.language_model, mock_lm)
        self.assertEqual(optimizer.mutation_temperature, 0.5)
        self.assertEqual(optimizer.crossover_temperature, 0.7)
        self.assertEqual(optimizer.selection, "best")
        self.assertEqual(optimizer.selection_temperature, 0.2)
        self.assertEqual(optimizer.merging_rate, 0.1)
        self.assertEqual(optimizer.population_size, 20)

    def test_invalid_selection(self):
        """Test that invalid selection raises ValueError."""
        with self.assertRaises(ValueError) as context:
            EvolutionaryOptimizer(selection="invalid")

        self.assertIn("selection", str(context.exception))

    def test_valid_selections(self):
        """Test that valid selections are accepted."""
        for selection in ["random", "best", "softmax"]:
            optimizer = EvolutionaryOptimizer(selection=selection)
            self.assertEqual(optimizer.selection, selection)

    def test_get_config(self):
        """Test that get_config returns all parameters."""
        optimizer = EvolutionaryOptimizer(
            mutation_temperature=0.4,
            crossover_temperature=0.6,
            selection="random",
            selection_temperature=0.25,
            merging_rate=0.05,
            population_size=15,
            name="config_test",
            description="Config test optimizer",
        )

        config = optimizer.get_config()

        self.assertEqual(config["mutation_temperature"], 0.4)
        self.assertEqual(config["crossover_temperature"], 0.6)
        self.assertEqual(config["selection"], "random")
        self.assertEqual(config["selection_temperature"], 0.25)
        self.assertEqual(config["merging_rate"], 0.05)
        self.assertEqual(config["population_size"], 15)

    async def test_on_batch_begin_epoch_zero(self):
        """Test on_batch_begin uses seed_candidates for epoch 0."""
        optimizer = EvolutionaryOptimizer(selection="random")

        seed_candidate = {"prompt": "seed_prompt"}
        trainable_variable = JsonDataModel(
            json={
                "seed_candidates": [seed_candidate],
                "best_candidates": [],
                "nb_visit": 5,
                "cumulative_reward": 1.0,
            },
            schema={
                "type": "object",
                "properties": {
                    "seed_candidates": {"type": "array"},
                    "best_candidates": {"type": "array"},
                    "nb_visit": {"type": "integer"},
                    "cumulative_reward": {"type": "number"},
                },
            },
        )

        await optimizer.on_batch_begin(0, 0, [trainable_variable])

        # Should reset nb_visit and cumulative_reward
        self.assertEqual(trainable_variable.get("nb_visit"), 0)
        self.assertEqual(trainable_variable.get("cumulative_reward"), 0.0)

    async def test_on_batch_begin_selection_best(self):
        """Test on_batch_begin with 'best' selection."""
        optimizer = EvolutionaryOptimizer(selection="best")

        best_candidates = [
            {"prompt": "low", "reward": 0.3},
            {"prompt": "high", "reward": 0.9},
            {"prompt": "mid", "reward": 0.6},
        ]

        trainable_variable = JsonDataModel(
            json={
                "seed_candidates": [],
                "best_candidates": best_candidates,
                "nb_visit": 0,
                "cumulative_reward": 0.0,
                "prompt": "initial",
            },
            schema={
                "type": "object",
                "properties": {
                    "seed_candidates": {"type": "array"},
                    "best_candidates": {"type": "array"},
                    "nb_visit": {"type": "integer"},
                    "cumulative_reward": {"type": "number"},
                    "prompt": {"type": "string"},
                },
            },
        )

        await optimizer.on_batch_begin(0, 1, [trainable_variable])

        # Should select the highest reward candidate
        self.assertEqual(trainable_variable.get("prompt"), "high")

    def test_select_candidate_empty(self):
        """Test select_candidate returns None for empty list."""
        optimizer = EvolutionaryOptimizer()
        result = optimizer.select_candidate([])
        self.assertIsNone(result)

    def test_select_candidate_random(self):
        """Test select_candidate with random selection."""
        optimizer = EvolutionaryOptimizer(selection="random")
        candidates = [
            {"prompt": "a", "reward": 0.5},
            {"prompt": "b", "reward": 0.8},
        ]
        result = optimizer.select_candidate(candidates)
        self.assertIn(result, candidates)

    def test_select_candidate_best(self):
        """Test select_candidate with best selection."""
        optimizer = EvolutionaryOptimizer(selection="best")
        candidates = [
            {"prompt": "low", "reward": 0.3},
            {"prompt": "high", "reward": 0.9},
            {"prompt": "mid", "reward": 0.6},
        ]
        result = optimizer.select_candidate(candidates)
        self.assertEqual(result["prompt"], "high")
        self.assertEqual(result["reward"], 0.9)

    def test_select_candidate_softmax(self):
        """Test select_candidate with softmax selection."""
        optimizer = EvolutionaryOptimizer(
            selection="softmax",
            selection_temperature=0.1,
        )
        candidates = [
            {"prompt": "low", "reward": 0.1},
            {"prompt": "high", "reward": 0.9},
        ]
        result = optimizer.select_candidate(candidates)
        self.assertIn(result, candidates)



from synalinks.src.api_export import synalinks_export
from synalinks.src.modules.core.action import Action
from synalinks.src.modules.core.branch import Branch
from synalinks.src.modules.core.decision import Decision
from synalinks.src.modules.core.generator import Generator
from synalinks.src.modules.core.generator import default_instructions
from synalinks.src.modules.core.generator import default_prompt_template
from synalinks.src.modules.core.identity import Identity
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.core.input_module import InputModule
from synalinks.src.modules.core.not_module import Not
from synalinks.src.modules.core.tool import Tool
from synalinks.src.modules.knowledge.embed_knowledge import EmbedKnowledge
from synalinks.src.modules.knowledge.retrieve_knowledge import RetrieveKnowledge
from synalinks.src.modules.knowledge.stamp_knowledge import StampKnowledge
from synalinks.src.modules.knowledge.update_knowledge import UpdateKnowledge
from synalinks.src.modules.module import Module
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought
from synalinks.src.modules.ttc.self_critique import SelfCritique
from synalinks.src.saving import serialization_lib


@synalinks_export("synalinks.modules.serialize")
def serialize(module):
    """Returns the module configuration as a Python dict.

    Args:
        module: A `synalinks.modules.module` instance to serialize.

    Returns:
        Python dict which contains the configuration of the module.
    """
    return serialization_lib.serialize_synalinks_object(module)


@synalinks_export("synalinks.modules.deserialize")
def deserialize(config, custom_objects=None):
    """Returns a Synalinks module object via its configuration.

    Args:
        config: A python dict containing a serialized module configuration.
        custom_objects: Optional dictionary mapping names (strings) to custom
            objects (classes and functions) to be considered during
            deserialization.

    Returns:
        A Synalinks module instance.
    """
    obj = serialization_lib.deserialize_synalinks_object(
        config,
        custom_objects=custom_objects,
    )
    if not isinstance(obj, Module):
        raise ValueError(
            "`synalinks.modules.deserialize` was passed a `config` object that is "
            f"not a `synalinks.modules.Module`. Received: {config}"
        )
    return obj



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought
from synalinks.src.modules.ttc.self_critique import SelfCritique
from synalinks.src.programs.program import Program


class SelfCritiqueModuleTest(testing.TestCase):
    @patch("litellm.acompletion")
    async def test_self_critique(self, mock_completion):
        class Query(DataModel):
            query: str = Field(
                description="The user query",
            )

        class Answer(DataModel):
            answer: str = Field(
                description="The correct answer",
            )

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        x0 = Input(data_model=Query)
        x1 = await ChainOfThought(
            data_model=Answer,
            language_model=language_model,
            return_inputs=True,
        )(x0)
        x2 = await SelfCritique(
            language_model=language_model,
        )(x1)

        program = Program(
            inputs=x0,
            outputs=x2,
            name="answer_with_cot_and_self_critique",
            description="Useful to answer accurately",
        )

        expected_answer = (
            """{"thinking": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "answer": "Toulouse"}"""
        )

        expected_critique = (
            """{"critique": "The response provided by the model is accurate and"""
            """ well-structured. The information about Toulouse's contribution"""
            """ to aerospace and robotics is also relevant. However, consider """
            """adding more conversational tone or humanizing the output slightly"""
            """ for better user experience.", "reward": 0.9}"""
        )

        mock_responses = [
            {"choices": [{"message": {"content": expected_answer}}]},
            {"choices": [{"message": {"content": expected_critique}}]},
        ]

        mock_completion.side_effect = mock_responses

        result = await program(
            Query(
                query="What is the French city of aerospace and robotics?",
            )
        )

        expected_string = (
            """{"query": "What is the French city of aerospace and robotics?","""
            """ "thinking": "Toulouse hosts numerous research institutions and """
            """universities that specialize in aerospace engineering and robotics,"""
            """ such as the Institut Supérieur de l'Aéronautique et de l'Espace """
            """(ISAE-SUPAERO) and the French National Centre for Scientific Research """
            """(CNRS)", "answer": "Toulouse", "critique": "The response provided by """
            """the model is accurate and well-structured. The information about """
            """Toulouse's contribution to aerospace and robotics is also relevant. """
            """However, consider adding more conversational tone or humanizing the """
            """output slightly for better user experience.", "reward": 0.9}"""
        )

        self.assertEqual(result.get_json(), json.loads(expected_string))



# Modified from: keras/src/backend/config.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import logging
import os
import random
import re

import numpy as np
import orjson

from synalinks.src.api_export import synalinks_export

# The type of float to use throughout a session.
_FLOATX = "float32"

# Epsilon fuzz factor used throughout the codebase.
_EPSILON = 1e-7

# The Synalinks API key.
_SYNALINKS_API_KEY = None

# Default backend: Pydantic.
_BACKEND = "pydantic"

# Default Synalinks Studio API base
_SYNALINKS_API_BASE = "http://localhost:8000/api"

# Enable observability
_ENABLE_OBSERVABILITY = False

# MLflow tracking URI for observability
_MLFLOW_TRACKING_URI = None

# MLflow experiment name for observability
_MLFLOW_EXPERIMENT_NAME = "synalinks_traces"

# Available backends
_AVAILABLE_BACKEND = ["pydantic"]

# Random seed for reproductability
_RANDOM_SEED = 123

np.random.seed(_RANDOM_SEED)
random.seed(_RANDOM_SEED)

logger = logging.getLogger("synalinks")
logger.propagate = False
logger.setLevel(logging.CRITICAL)


class SynalinksLogFormatter(logging.Formatter):
    """Formatter for console logging with colors and prefix."""

    blue = "\x1b[34m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    prefix = "🧠🔗 Synalinks: "

    FORMATS = {
        logging.DEBUG: f"(DEBUG) {prefix}%(message)s",
        logging.INFO: f"{prefix}%(message)s",
        logging.WARNING: f"{prefix}%(message)s",
        logging.ERROR: f"{bold_red}{prefix}%(message)s{reset}",
        logging.CRITICAL: f"{bold_red}{prefix}%(message)s{reset}",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno, self.FORMATS[logging.INFO])
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class SynalinksFileFormatter(logging.Formatter):
    """Formatter for file logging that removes ANSI escape codes."""

    ANSI_ESCAPE_PATTERN = re.compile(r"\x1b\[[0-9;]*m")

    def format(self, record):
        record.msg = self.ANSI_ESCAPE_PATTERN.sub("", str(record.msg))
        return super().format(record)


_ENABLE_TELEMETRY = True


@synalinks_export(
    [
        "synalinks.config.disable_telemetry",
        "synalinks.backend.disable_telemetry",
        "synalinks.disable_telemetry",
    ]
)
def disable_telemetry():
    global _ENABLE_TELEMETRY
    _ENABLE_TELEMETRY = False


@synalinks_export(
    [
        "synalinks.config.is_telemetry_enabled",
        "synalinks.backend.is_telemetry_enabled",
        "synalinks.is_telemetry_enabled",
    ]
)
def is_telemetry_enabled():
    return _ENABLE_TELEMETRY


@synalinks_export(["synalinks.config.floatx", "synalinks.backend.floatx"])
def floatx():
    """Return the default float type, as a string.

    E.g. `'bfloat16'`, `'float16'`, `'float32'`, `'float64'`.

    Returns:
        (str): The current default float type.

    Example:

    ```python
    >>> synalinks.config.floatx()
    'float32'
    ```

    """
    return _FLOATX


@synalinks_export(["synalinks.config.set_floatx", "synalinks.backend.set_floatx"])
def set_floatx(value):
    """Set the default float dtype.

    Note: It is not recommended to set this to `"float16"`,
    as this will likely cause numeric stability issues.
    Instead, use `float64` or `float32`.

    Args:
        value (str): The float type between `'bfloat16'`, `'float16'`, `'float32'`,
            or `'float64'`.

    Examples:

    ```python
    >>> synalinks.config.floatx()
    'float32'
    ```

    ```python
    >>> synalinks.config.set_floatx('float64')
    >>> synalinks.config.floatx()
    'float64'
    ```

    ```python
    >>> # Set it back to float32
    >>> synalinks.config.set_floatx('float32')
    ```

    Raises:
        ValueError: In case of invalid value.
    """
    global _FLOATX
    accepted_dtypes = {"bfloat16", "float16", "float32", "float64"}
    if value not in accepted_dtypes:
        raise ValueError(
            f"Unknown `floatx` value: {value}. Expected one of {accepted_dtypes}"
        )
    _FLOATX = str(value)


@synalinks_export(["synalinks.config.epsilon", "synalinks.backend.epsilon"])
def epsilon():
    """Return the value of the fuzz factor used in numeric expressions.

    Returns:
        (float): The epsilon value.

    Example:

    ```python
    >>> synalinks.config.epsilon()
    1e-07
    ```

    """
    return _EPSILON


@synalinks_export(["synalinks.config.set_epsilon", "synalinks.backend.set_epsilon"])
def set_epsilon(value):
    """Set the value of the fuzz factor used in numeric expressions.

    Args:
        value (float): The new value of epsilon.

    Examples:

    ```python
    >>> synalinks.config.epsilon()
    1e-07
    ```

    ```python
    >>> synalinks.config.set_epsilon(1e-5)
    >>> synalinks.config.epsilon()
    1e-05
    ```

    ```python
    >>> # Set it back to the default value.
    >>> synalinks.config.set_epsilon(1e-7)
    ```

    """
    global _EPSILON
    _EPSILON = value


@synalinks_export(
    [
        "synalinks.config.set_seed",
        "synalinks.backend.set_seed",
        "synalinks.set_seed",
    ]
)
def set_seed(value):
    """Set the value of the random seed for reproductability.

    Args:
        value (float): The new value of epsilon.
    """
    global _RANDOM_SEED
    _RANDOM_SEED = value
    np.random.seed(_RANDOM_SEED)
    random.seed(_RANDOM_SEED)


@synalinks_export(
    [
        "synalinks.config.get_seed",
        "synalinks.backend.get_seed",
        "synalinks.get_seed",
    ]
)
def get_seed():
    return _RANDOM_SEED


@synalinks_export(
    [
        "synalinks.config.enable_logging",
        "synalinks.backend.enable_logging",
        "synalinks.enable_logging",
    ]
)
def enable_logging(log_level="debug", log_to_file=False):
    """
    Configures and enables logging for the application.

    This function sets up the logging configuration for the application,
    allowing logs to be output either to a specified file or to the console.
    The logging level can be set to DEBUG for more verbose logging or INFO
    for standard logging.

    Args:
        log_level (str): The logging level.
        log_to_file (bool): If True save the logs into `synalinks.log`

    The log message format includes the timestamp, log level, and the log message itself.
    """
    global logger

    log_level = log_level.upper()
    if log_level and hasattr(logging, log_level):
        log_level = getattr(logging, log_level)

    logger.setLevel(log_level)

    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(log_level)
    stream_handler.setFormatter(SynalinksLogFormatter())
    logger.addHandler(stream_handler)

    if log_to_file:
        file_handler = logging.FileHandler("synalinks.log", mode="w")
        file_handler.setLevel(log_level)
        formatter = SynalinksLogFormatter("%(asctime)s - %(levelname)s - %(message)s")
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)


@synalinks_export(
    [
        "synalinks.config.is_observability_enabled",
        "synalinks.backend.is_observability_enabled",
        "synalinks.is_observability_enabled",
    ]
)
def is_observability_enabled():
    """Check if the observability is enabled

    Returns:
        (bool): True if the observability is enabled.
    """
    return _ENABLE_OBSERVABILITY


@synalinks_export(
    [
        "synalinks.config.api_base",
        "synalinks.backend.api_base",
        "synalinks.api_base",
    ]
)
def api_base():
    """Returns the Synalinks api base

    Returns:
        (str): The observability api base
    """
    return _SYNALINKS_API_BASE


@synalinks_export(
    [
        "synalinks.config.set_api_base",
        "synalinks.backend.set_api_base",
        "synalinks.set_api_base",
    ]
)
def set_api_base(api_base):
    """Set the observability api base

    Args:
        api_base (str): The observability api base
    """
    global _SYNALINKS_API_BASE
    _SYNALINKS_API_BASE = api_base


@synalinks_export(
    [
        "synalinks.config.enable_observability",
        "synalinks.backend.enable_observability",
        "synalinks.enable_observability",
    ]
)
def enable_observability(tracking_uri=None, experiment_name=None):
    """
    Configures and enables observability for the application using MLflow.

    This function sets up the observability configuration for the application,
    enabling tracing of module calls via MLflow.

    Args:
        tracking_uri (str): Optional. The MLflow tracking server URI.
            If not provided, MLflow will use the default (local ./mlruns
            directory or MLFLOW_TRACKING_URI environment variable).
        experiment_name (str): Optional. The MLflow experiment name.
            Defaults to "synalinks_traces".

    Example:

    ```python
    import synalinks

    # Basic usage with defaults
    synalinks.enable_observability()

    # With custom MLflow tracking server
    synalinks.enable_observability(
        tracking_uri="http://localhost:5000",
        experiment_name="my_experiment"
    )
    ```
    """
    global _MLFLOW_TRACKING_URI
    global _MLFLOW_EXPERIMENT_NAME
    global _ENABLE_OBSERVABILITY

    if tracking_uri:
        _MLFLOW_TRACKING_URI = tracking_uri
    if experiment_name:
        _MLFLOW_EXPERIMENT_NAME = experiment_name
    _ENABLE_OBSERVABILITY = True


@synalinks_export(
    [
        "synalinks.config.mlflow_tracking_uri",
        "synalinks.backend.mlflow_tracking_uri",
    ]
)
def mlflow_tracking_uri():
    """Returns the MLflow tracking URI for observability.

    Returns:
        (str): The MLflow tracking URI, or None if not set.
    """
    return _MLFLOW_TRACKING_URI


@synalinks_export(
    [
        "synalinks.config.mlflow_experiment_name",
        "synalinks.backend.mlflow_experiment_name",
    ]
)
def mlflow_experiment_name():
    """Returns the MLflow experiment name for observability.

    Returns:
        (str): The MLflow experiment name.
    """
    return _MLFLOW_EXPERIMENT_NAME


@synalinks_export(
    [
        "synalinks.config.api_key",
        "synalinks.backend.api_key",
    ]
)
def api_key():
    """Synalinks API key.

    Returns:
        (str): Synalinks API key.

    ```python
    >>> synalinks.config.api_key()
    'my-secret-api-key'
    ```

    """
    return _SYNALINKS_API_KEY


@synalinks_export(
    [
        "synalinks.config.set_api_key",
        "synalinks.backend.set_api_key",
    ]
)
def set_api_key(key):
    """Set Synalinks API key.

    Args:
        key (str): The API key value.

    The API key is retrieved from the env variables at start.

    ```python
    >>> os.environ["SYNALINKS_API_KEY"] = 'my-secret-api-key'
    ```

    Or you can setup it using the config

    ```python
    >>> synalinks.config.set_api_key('my-secret-api-key')
    >>> synalinks.config.api_key()
    'my-secret-api-key'
    ```

    Args:
        key (str): Synalinks API key.
    """
    global _SYNALINKS_API_KEY
    _SYNALINKS_API_KEY = key


# Set synalinks base dir path given synalinks_HOME env variable, if applicable.
# Otherwise either ~/.synalinks or /tmp.
if "SYNALINKS_HOME" in os.environ:
    _synalinks_DIR = os.environ.get("SYNALINKS_HOME")
else:
    _synalinks_base_dir = os.path.expanduser("~")
    if not os.access(_synalinks_base_dir, os.W_OK):
        _synalinks_base_dir = "/tmp"
    _synalinks_DIR = os.path.join(_synalinks_base_dir, ".synalinks")


@synalinks_export(["synalinks.config.synalinks_home", "synalinks.synalinks_home"])
def synalinks_home():
    # Private accessor for the synalinks home location.
    return _synalinks_DIR


@synalinks_export(["synalinks.config.backend", "synalinks.backend.backend"])
def backend():
    """Publicly accessible method for determining the current backend.

    Returns:
        (str): The name of the backend synalinks is currently using. like
            `"pydantic"`.

    Example:

    ```python
    >>> synalinks.config.backend()
    'pydantic'
    ```
    """
    return _BACKEND


@synalinks_export(["synalinks.config.set_backend", "synalinks.backend.set_backend"])
def set_backend(backend):
    if backend not in _AVAILABLE_BACKEND:
        raise ValueError(
            f"Cannot set backend to {backend}, "
            f"you should choose between: {_AVAILABLE_BACKEND}"
        )
    global _BACKEND
    _BACKEND = backend


# Attemp to get the API key from env variables
_api_key = os.getenv("SYNALINKS_API_KEY", None)
if _api_key:
    assert isinstance(_api_key, str)
    set_api_key(_api_key)

# Attempt to read synalinks config file.
_config_path = os.path.expanduser(os.path.join(_synalinks_DIR, "synalinks.json"))
if os.path.exists(_config_path):
    try:
        with open(_config_path, "rb") as f:
            _config = orjson.loads(f.read())
    except ValueError:
        _config = {}
    _backend = _config.get("backend", _BACKEND)
    assert _backend in _AVAILABLE_BACKEND
    _floatx = _config.get("floatx", _FLOATX)
    assert _floatx in {"float16", "float32", "float64"}
    _epsilon = _config.get("epsilon", _EPSILON)
    assert isinstance(_epsilon, float)
    _seed = _config.get("seed", _RANDOM_SEED)
    assert isinstance(_seed, int)
    _SYNALINKS_API_BASE = _config.get("api_base", _SYNALINKS_API_BASE)
    assert isinstance(_SYNALINKS_API_BASE, str)

    set_backend(_backend)
    set_floatx(_floatx)
    set_epsilon(_epsilon)
    set_seed(_seed)
    set_api_base(_SYNALINKS_API_BASE)

# Save config file, if possible.
if not os.path.exists(_synalinks_DIR):
    try:
        os.makedirs(_synalinks_DIR)
    except OSError:
        # Except permission denied and potential race conditions
        # in multi-threaded environments.
        pass

if not os.path.exists(_config_path):
    _config = {
        "backend": _BACKEND,
        "floatx": _FLOATX,
        "epsilon": _EPSILON,
        "seed": _RANDOM_SEED,
        "api_base": _SYNALINKS_API_BASE,
    }
    try:
        with open(_config_path, "wb") as f:
            f.write(orjson.dumps(_config, option=orjson.OPT_INDENT_2))
    except IOError:
        # Except permission denied.
        pass



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.utils.nlp_utils import add_suffix
from synalinks.src.utils.nlp_utils import is_plural
from synalinks.src.utils.nlp_utils import normalize_and_tokenize
from synalinks.src.utils.nlp_utils import normalize_text
from synalinks.src.utils.nlp_utils import remove_articles
from synalinks.src.utils.nlp_utils import remove_numerical_suffix
from synalinks.src.utils.nlp_utils import remove_punctuation
from synalinks.src.utils.nlp_utils import to_plural
from synalinks.src.utils.nlp_utils import to_plural_property
from synalinks.src.utils.nlp_utils import to_plural_without_numerical_suffix
from synalinks.src.utils.nlp_utils import to_singular
from synalinks.src.utils.nlp_utils import to_singular_property
from synalinks.src.utils.nlp_utils import to_singular_without_numerical_suffix


class NLPUtilsTest(testing.TestCase):
    def test_to_plural(self):
        # Test regular plurals
        self.assertEqual(to_plural("cat"), "cats")
        self.assertEqual(to_plural("dog"), "dogs")
        self.assertEqual(to_plural("car"), "cars")

        # Test words ending in 'y'
        self.assertEqual(to_plural("city"), "cities")
        self.assertEqual(to_plural("baby"), "babies")

        # Test words ending in 's', 'sh', 'ch', 'x', 'z'
        self.assertEqual(to_plural("bus"), "buses")
        self.assertEqual(to_plural("watch"), "watches")
        self.assertEqual(to_plural("box"), "boxes")
        self.assertEqual(to_plural("quiz"), "quizzes")

        # Test irregular plurals
        self.assertEqual(to_plural("child"), "children")
        self.assertEqual(to_plural("man"), "men")
        self.assertEqual(to_plural("person"), "people")
        self.assertEqual(to_plural("mouse"), "mice")
        self.assertEqual(to_plural("tooth"), "teeth")
        self.assertEqual(to_plural("foot"), "feet")

        # Test words that are the same in singular and plural
        self.assertEqual(to_plural("fish"), "fish")
        self.assertEqual(to_plural("sheep"), "sheep")
        self.assertEqual(to_plural("series"), "series")

        # Test words not in the dictionary
        self.assertEqual(to_plural("robot"), "robots")
        self.assertEqual(to_plural("algorithm"), "algorithms")

    def test_to_singular(self):
        # Test regular singulars
        self.assertEqual(to_singular("cats"), "cat")
        self.assertEqual(to_singular("dogs"), "dog")
        self.assertEqual(to_singular("cars"), "car")

        # Test words ending in 'ies'
        self.assertEqual(to_singular("cities"), "city")
        self.assertEqual(to_singular("babies"), "baby")

        # Test words ending in 'es'
        self.assertEqual(to_singular("buses"), "bus")
        self.assertEqual(to_singular("watches"), "watch")
        self.assertEqual(to_singular("boxes"), "box")
        self.assertEqual(to_singular("quizzes"), "quiz")

        # Test irregular singulars
        self.assertEqual(to_singular("children"), "child")
        self.assertEqual(to_singular("men"), "man")
        self.assertEqual(to_singular("people"), "person")
        self.assertEqual(to_singular("mice"), "mouse")
        self.assertEqual(to_singular("teeth"), "tooth")
        self.assertEqual(to_singular("feet"), "foot")

        # Test words that are the same in singular and plural
        self.assertEqual(to_singular("fish"), "fish")
        self.assertEqual(to_singular("sheep"), "sheep")
        self.assertEqual(to_singular("series"), "series")

        # Test words not in the dictionary
        self.assertEqual(to_singular("robots"), "robot")
        self.assertEqual(to_singular("algorithms"), "algorithm")

    def test_to_plural_property(self):
        self.assertEqual(to_plural_property("user_account"), "user_accounts")
        self.assertEqual(to_plural_property("product_category"), "product_categories")
        self.assertEqual(to_plural_property("city"), "cities")

    def test_to_singular_property(self):
        self.assertEqual(to_singular_property("user_accounts"), "user_account")
        self.assertEqual(to_singular_property("product_categories"), "product_category")
        self.assertEqual(to_singular_property("cities"), "city")

    def test_remove_numerical_suffix(self):
        self.assertEqual(remove_numerical_suffix("property_key_1"), "property_key")
        self.assertEqual(remove_numerical_suffix("property_key_123"), "property_key")
        self.assertEqual(remove_numerical_suffix("property_key"), "property_key")

    def test_add_suffix(self):
        self.assertEqual(add_suffix("property_key", 1), "property_key_1")
        self.assertEqual(add_suffix("property_key", 123), "property_key_123")

    def test_to_singular_without_numerical_suffix(self):
        self.assertEqual(
            to_singular_without_numerical_suffix("property_key_1"), "property_key"
        )
        self.assertEqual(
            to_singular_without_numerical_suffix("user_accounts_2"), "user_account"
        )

    def test_to_plural_without_numerical_suffix(self):
        self.assertEqual(
            to_plural_without_numerical_suffix("property_key_1"), "property_keys"
        )
        self.assertEqual(
            to_plural_without_numerical_suffix("user_account_2"), "user_accounts"
        )

    def test_is_plural(self):
        self.assertTrue(is_plural("cities"))
        self.assertFalse(is_plural("city"))
        self.assertTrue(is_plural("user_accounts"))
        self.assertFalse(is_plural("user_account"))

    def test_remove_articles(self):
        self.assertEqual(remove_articles("the quick brown fox"), "quick brown fox")
        self.assertEqual(remove_articles("an apple a day"), "apple day")

    def test_remove_punctuation(self):
        self.assertEqual(remove_punctuation("Hello, world!"), "Hello world")
        self.assertEqual(remove_punctuation("Test: 1, 2, 3..."), "Test 1 2 3")

    def test_normalize_text(self):
        self.assertEqual(normalize_text("The Quick Brown Fox!"), "quick brown fox")
        self.assertEqual(normalize_text("An Apple a Day..."), "apple day")

    def test_normalize_and_tokenize(self):
        self.assertEqual(
            normalize_and_tokenize("The Quick Brown Fox!"), ["quick", "brown", "fox"]
        )
        self.assertEqual(normalize_and_tokenize("An Apple a Day..."), ["apple", "day"])
